from __future__ import annotations

import sys
from typing import Any, ChainMap, Dict, MutableMapping, Optional
from weakref import finalize

from typing_extensions import Final

from ..events.event import Event, EventEmitter
from ..translations import trans

_null = object()


class Context(ChainMap):
    def __init__(self, *maps: Dict[str, Any]) -> None:
        super().__init__(*maps)
        self.changed = EventEmitter(self, 'changed')

    # this requires psygnal or changes to events.py ...
    # @contextmanager
    # def buffered_changes(self):
    #     with self.changed.paused(lambda a, b: (a[0].union(b[0]),)):
    #         yield

    def __setitem__(self, k: str, v: Any) -> None:
        emit = self.get(k, _null) is not v
        super().__setitem__(k, v)
        if emit:
            self.changed(value={k})

    def __delitem__(self, k: str) -> None:
        emit = k in self
        super().__delitem__(k)
        if emit:
            self.changed(value={k})

    def new_child(self, m: Optional[MutableMapping] = None) -> Context:
        new = super().new_child(m=m)
        self.changed.connect(new.changed)
        return new

    def __hash__(self):
        return id(self)


class SettingsAwareContext(Context):
    """A special context that allows access of settings using `settings.`

    This takes no parents, and will always be a root context.
    """

    _PREFIX: Final[str] = 'settings.'

    def __init__(self):
        super().__init__()
        from ...settings import get_settings

        self._settings = get_settings()
        self._settings.events.changed.connect(self._update_key)

    def _update_key(self, event: Event):
        self.changed(value={f'{self._PREFIX}{event.key}'})

    def __del__(self):
        self._settings.events.changed.disconnect(self._update_key)

    def __missing__(self, key: str) -> Any:
        if key.startswith(self._PREFIX):
            splits = [k for k in key.split(".")[1:] if k]
            val: Any = self._settings
            if splits:
                while splits:
                    val = getattr(val, splits.pop(0))
                if hasattr(val, 'dict'):
                    val = val.dict()
                return val
        return super().__missing__(key)

    def new_child(self, m: Optional[dict] = None) -> Context:  # type: ignore
        """New ChainMap with a new map followed by all previous maps.

        If no map is provided, an empty dict is used.
        """
        # important to use self, not *self.maps
        return Context(m or {}, self)  # type: ignore

    def __setitem__(self, k: str, v: Any) -> None:
        if k.startswith(self._PREFIX):
            raise ValueError(
                trans._(
                    "Cannot set key starting with {prefix!r}",
                    deferred=True,
                    prefix=self._PREFIX,
                )
            )

        return super().__setitem__(k, v)

    def __bool__(self):
        return True


_ROOT_CONTEXT: Optional[SettingsAwareContext] = None

# note: it seems like WeakKeyDictionary would be a nice match here, but
# it appears that the object somehow isn't initialized "enough" to register
# as the same object in the WeakKeyDictionary later when queried with
# `obj in _OBJ_TO_CONTEXT` ... so instead we use id(obj)
# _OBJ_TO_CONTEXT: WeakKeyDictionary[object, Context] = WeakKeyDictionary()
_OBJ_TO_CONTEXT: Dict[int, Context] = {}


def create_context(
    obj: object,
    max_depth: int = 20,
    start: int = 2,
    root: Optional[Context] = None,
) -> Optional[Context]:

    if root is None:
        global _ROOT_CONTEXT
        if _ROOT_CONTEXT is None:
            _ROOT_CONTEXT = SettingsAwareContext()
        root = _ROOT_CONTEXT
    else:
        assert isinstance(root, Context), trans._(
            'root must be an instance of Context', deferred=True
        )

    parent = root
    if hasattr(sys, '_getframe'):  # CPython implementation detail
        frame = sys._getframe(start)
        i = 0
        # traverse call stack looking for another object that has a context
        # to scope this new context off of.
        while frame and i < max_depth:
            if frame.f_code.co_name in (
                '__new__',
                '_set_default_and_type',
            ):
                # type is being declared and pydantic is checking defaults
                # this context will never be used.
                return None
            elif 'self' in frame.f_locals:
                _ctx = _OBJ_TO_CONTEXT.get(id(frame.f_locals['self']))
                if _ctx is not None:
                    parent = _ctx
                    break
            frame = frame.f_back  # type: ignore
            i += 1

    new_context = parent.new_child()
    obj_id = id(obj)
    _OBJ_TO_CONTEXT[obj_id] = new_context
    # remove key from dict when object is deleted
    finalize(obj, lambda: _OBJ_TO_CONTEXT.pop(obj_id, None))
    return new_context


def get_context(obj: object) -> Optional[Context]:
    return _OBJ_TO_CONTEXT.get(id(obj))



import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLabel, QPushButton

from napari._qt.widgets.qt_extension2reader import Extension2ReaderTable
from napari._tests.utils import restore_settings_on_exit
from napari.settings import get_settings


@pytest.fixture
def extension2reader_widget(qtbot):
    def _extension2reader_widget(**kwargs):
        widget = Extension2ReaderTable(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _extension2reader_widget


def test_extension2reader_defaults(
    extension2reader_widget,
):
    with restore_settings_on_exit():
        get_settings().plugins.extension2reader = {}
        widget = extension2reader_widget()

        assert widget._table.rowCount() == 1
        assert widget._table.itemAt(0, 0).text() == 'No extensions found.'


def test_extension2reader_with_settings(
    extension2reader_widget,
):
    with restore_settings_on_exit():
        get_settings().plugins.extension2reader = {'.test': 'test-plugin'}
        widget = extension2reader_widget()

        assert widget._table.rowCount() == 1
        assert widget._table.item(0, 0).text() == '.test'
        assert (
            widget._table.cellWidget(0, 1).findChild(QLabel).text()
            == 'test-plugin'
        )


def test_extension2reader_removal(extension2reader_widget, qtbot):
    with restore_settings_on_exit():
        get_settings().plugins.extension2reader = {
            '.test': 'test-plugin',
            '.abc': 'abc-plugin',
        }
        widget = extension2reader_widget()

        assert widget._table.rowCount() == 2

        btn_to_click = widget._table.cellWidget(0, 1).findChild(QPushButton)
        qtbot.mouseClick(btn_to_click, Qt.LeftButton)

        assert get_settings().plugins.extension2reader == {
            '.abc': 'abc-plugin'
        }
        assert widget._table.rowCount() == 1
        assert widget._table.item(0, 0).text() == '.abc'



import pytest

from napari._qt.widgets.qt_range_slider_popup import QRangeSliderPopup

initial = (100, 400)
range_ = (0, 500)


@pytest.fixture
def popup(qtbot):
    popup = QRangeSliderPopup()
    popup.slider.setRange(*range_)
    popup.slider.setValue(initial)
    qtbot.addWidget(popup)
    return popup


def test_range_slider_popup_labels(popup):
    """make sure labels are correct"""
    assert popup.slider._handle_labels[0].value() == initial[0]
    assert popup.slider._handle_labels[1].value() == initial[1]
    assert (popup.slider.minimum(), popup.slider.maximum()) == range_


def test_range_slider_changes_labels(popup):
    """make sure setting the slider updates the labels"""
    popup.slider.setValue((10, 20))
    assert popup.slider._handle_labels[0].value() == 10
    assert popup.slider._handle_labels[1].value() == 20



from typing import Union

from ...utils.events import EventedModel
from ...utils.events.custom_types import Array


class SurfaceWireframe(EventedModel):
    """
    Wireframe representation of the edges of a surface mesh.

    Attributes
    ----------
    visible : bool
        Whether the wireframe is displayed.
    color : str, array-like
        If string can be any color name recognized by vispy or hex value if
        starting with `#`. If array-like must be 1-dimensional array with 3
        or 4 elements.
    width : float
        The width of the wireframe lines.
    """

    visible: bool = False
    color: Union[str, Array[float, (3,)], Array[float, (4,)]] = 'black'
    width: float = 1



import gc
import sys
from unittest import mock

import pytest

import napari
from napari import __main__


@pytest.fixture
def mock_run():
    """mock to prevent starting the event loop."""
    with mock.patch('napari._qt.widgets.qt_splash_screen.NapariSplashScreen'):
        with mock.patch('napari.run'):
            yield napari.run


def test_cli_works(monkeypatch, capsys):
    """Test the cli runs and shows help"""
    monkeypatch.setattr(sys, 'argv', ['napari', '-h'])
    with pytest.raises(SystemExit):
        __main__._run()
    assert 'napari command line viewer.' in str(capsys.readouterr())


def test_cli_shows_plugins(napari_plugin_manager, monkeypatch, capsys):
    """Test the cli --info runs and shows plugins"""
    monkeypatch.setattr(sys, 'argv', ['napari', '--info'])
    with pytest.raises(SystemExit):
        __main__._run()
    # this is because sckit-image is OUR builtin providing sample_data
    assert 'scikit-image' in str(capsys.readouterr())


def test_cli_parses_unknowns(mock_run, monkeypatch):
    """test that we can parse layer keyword arg variants"""

    def assert_kwargs(*args, **kwargs):
        assert args == (["file"],)
        assert kwargs['contrast_limits'] == (0, 1)

    # testing all the variants of literal_evals
    monkeypatch.setattr(napari, 'view_path', assert_kwargs)
    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['n', 'file', '--contrast-limits', '(0, 1)'])
        __main__._run()
    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['n', 'file', '--contrast-limits', '(0,1)'])
        __main__._run()
    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['n', 'file', '--contrast-limits=(0, 1)'])
        __main__._run()
    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['n', 'file', '--contrast-limits=(0,1)'])
        __main__._run()


def test_cli_raises(monkeypatch):
    """test that unknown kwargs raise the correct errors."""
    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['napari', 'path/to/file', '--nonsense'])
        with pytest.raises(SystemExit) as e:
            __main__._run()
        assert str(e.value) == 'error: unrecognized arguments: --nonsense'

    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['napari', 'path/to/file', '--gamma'])
        with pytest.raises(SystemExit) as e:
            __main__._run()
        assert str(e.value) == 'error: argument --gamma expected one argument'


@mock.patch('runpy.run_path')
def test_cli_runscript(run_path, monkeypatch, tmp_path):
    """Test that running napari script.py runs a script"""
    script = tmp_path / 'test.py'
    script.write_text('import napari; v = napari.Viewer(show=False)')

    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['napari', str(script)])
        __main__._run()

    run_path.assert_called_once_with(str(script))


@mock.patch('napari.view_path')
def test_cli_passes_kwargs(view_path, mock_run, monkeypatch):
    """test that we can parse layer keyword arg variants"""

    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['n', 'file', '--name', 'some name'])
        __main__._run()

    view_path.assert_called_once_with(
        ['file'],
        stack=False,
        plugin=None,
        layer_type=None,
        name='some name',
    )
    mock_run.assert_called_once_with(gui_exceptions=True)


def test_cli_retains_viewer_ref(mock_run, monkeypatch, make_napari_viewer):
    """Test that napari.__main__ is retaining a reference to the viewer."""
    v = make_napari_viewer()  # our mock view_path will return this object
    ref_count = None  # counter that will be updated before __main__._run()

    def _check_refs(**kwargs):
        # when run() is called in napari.__main__, we will call this function
        # it forces garbage collection, and then makes sure that at least one
        # additional reference to our viewer exists.
        gc.collect()
        if sys.getrefcount(v) <= ref_count:
            raise AssertionError(
                "Reference to napari.viewer has been lost by "
                "the time the event loop started in napari.__main__"
            )

    mock_run.side_effect = _check_refs
    with monkeypatch.context() as m:
        m.setattr(sys, 'argv', ['napari', 'path/to/file.tif'])
        # return our local v
        with mock.patch('napari.view_path', return_value=v) as mock_vp:
            ref_count = sys.getrefcount(v)  # count current references
            __main__._run()
            mock_vp.assert_called_once()



import numpy as np

from napari._qt.widgets.qt_action_context_menu import QtActionContextMenu
from napari.components.layerlist import LayerList
from napari.layers import Image, Labels
from napari.layers._layer_actions import _LAYER_ACTIONS
from napari.utils.context._expressions import parse_expression


def test_action_menu(qapp):
    ACTIONS = {
        'add_one': {
            'description': 'Add one',
            'action': lambda x: x.append(1),
            'enable_when': parse_expression('count == 0'),
        }
    }

    menu = QtActionContextMenu(ACTIONS)
    menu.update_from_context({'count': 0, 'is_ready': True})
    assert menu._get_action('add_one').isEnabled()
    assert menu._get_action('add_one').isVisible()

    CONTEXT_KEYS = {
        'count': lambda x: len(x),
        'is_ready': lambda x: True,
    }
    some_object = [42]
    ctx = {k: v(some_object) for k, v in CONTEXT_KEYS.items()}
    menu.update_from_context(ctx)
    assert not menu._get_action('add_one').isEnabled()


def test_layer_action_menu(qapp):
    """Test the actions in LAYER_ACTIONS."""
    menu = QtActionContextMenu(_LAYER_ACTIONS)
    layer_list = LayerList([])
    menu.update_from_context(layer_list._ctx)
    assert not menu._get_action('napari:convert_to_image').isEnabled()

    layer_list.append(Labels(np.zeros((8, 8), int)))
    menu.update_from_context(layer_list._ctx)
    assert menu._get_action('napari:convert_to_image').isEnabled()
    assert not menu._get_action('napari:convert_to_labels').isEnabled()

    layer_list.append(Image(np.zeros((8, 8))))
    menu.update_from_context(layer_list._ctx)
    assert not menu._get_action('napari:convert_to_image').isEnabled()
    assert menu._get_action('napari:convert_to_labels').isEnabled()
    assert not menu._get_action('napari:link_selected_layers').isEnabled()

    layer_list.select_all()
    menu.update_from_context(layer_list._ctx)
    assert menu._get_action('napari:link_selected_layers').isEnabled()
    assert not menu._get_action('napari:unlink_selected_layers').isEnabled()

    layer_list.link_layers()
    menu.update_from_context(layer_list._ctx)
    assert not menu._get_action('napari:link_selected_layers').isEnabled()
    assert menu._get_action('napari:unlink_selected_layers').isEnabled()



from typing import TYPE_CHECKING

from ...utils.translations import trans
from ._util import NapariMenu, populate_menu

if TYPE_CHECKING:
    from ..qt_main_window import Window


class WindowMenu(NapariMenu):
    def __init__(self, window: 'Window'):
        super().__init__(trans._('&Window'), window._qt_window)
        ACTIONS = []
        populate_menu(self, ACTIONS)



import numpy as np

from ..visuals.vectors import VectorsVisual
from .base import VispyBaseLayer


class VispyVectorsLayer(VispyBaseLayer):
    def __init__(self, layer):
        node = VectorsVisual()
        super().__init__(layer, node)

        self.layer.events.edge_color.connect(self._on_data_change)

        self.reset()
        self._on_data_change()

    def _on_data_change(self):
        if (
            len(self.layer._view_vertices) == 0
            or len(self.layer._view_faces) == 0
        ):
            vertices = np.zeros((3, self.layer._ndisplay))
            faces = np.array([[0, 1, 2]])
            face_color = np.array([[0, 0, 0, 0]])
        else:
            vertices = self.layer._view_vertices[:, ::-1]
            faces = self.layer._view_faces
            face_color = self.layer._view_face_color

        if self.layer._ndisplay == 3 and self.layer.ndim == 2:
            vertices = np.pad(vertices, ((0, 0), (0, 1)), mode='constant')

        # self.node.set_data(
        #     vertices=vertices, faces=faces, color=self.layer.current_edge_color
        # )
        self.node.set_data(
            vertices=vertices,
            faces=faces,
            face_colors=face_color,
        )

        self.node.update()
        # Call to update order of translation values with new dims:
        self._on_matrix_change()



import numpy as np
import pytest

from napari.layers import Image
from napari.layers.utils.interactivity_utils import (
    drag_data_to_projected_distance,
    orient_plane_normal_around_cursor,
)


@pytest.mark.parametrize(
    "start_position, end_position, view_direction, vector, expected_value",
    [
        # drag vector parallel to view direction
        # projected onto perpendicular vector
        ([0, 0, 0], [0, 0, 1], [0, 0, 1], [1, 0, 0], 0),
        # same as above, projection onto multiple perpendicular vectors
        # should produce multiple results
        ([0, 0, 0], [0, 0, 1], [0, 0, 1], [[1, 0, 0], [0, 1, 0]], [0, 0]),
        # drag vector perpendicular to view direction
        # projected onto itself
        ([0, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 0], 1),
        # drag vector perpendicular to view direction
        # projected onto itself
        ([0, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 0], 1),
    ],
)
def test_drag_data_to_projected_distance(
    start_position, end_position, view_direction, vector, expected_value
):
    result = drag_data_to_projected_distance(
        start_position, end_position, view_direction, vector
    )
    assert np.allclose(result, expected_value)


@pytest.mark.parametrize(
    'layer',
    [
        Image(np.zeros(shape=(28, 28, 28))),
        Image(np.zeros(shape=(2, 28, 28, 28))),
    ],
)
def test_orient_plane_normal_around_cursor(make_napari_viewer, layer):
    viewer = make_napari_viewer()
    viewer.dims.ndisplay = 3
    viewer.camera.angles = (0, 0, 90)
    viewer.cursor.position = [14] * layer._ndim

    viewer.add_layer(layer)
    layer.depiction = 'plane'
    layer.plane.normal = (1, 0, 0)
    layer.plane.position = (14, 14, 14)

    # apply simple transformation on the volume
    layer.translate = [1] * layer._ndim

    # orient plane normal
    orient_plane_normal_around_cursor(layer=layer, plane_normal=(1, 0, 1))

    # check that plane normal has been updated
    assert np.allclose(
        layer.plane.normal, [1, 0, 1] / np.linalg.norm([1, 0, 1])
    )
    assert np.allclose(layer.plane.position, (14, 13, 13))



from contextlib import nullcontext

import dask
import dask.array as da
import numpy as np
import pytest

from napari import layers
from napari.components import ViewerModel
from napari.utils import _dask_utils, resize_dask_cache


@pytest.mark.sync_only
@pytest.mark.parametrize('dtype', ['float64', 'uint8'])
def test_dask_not_greedy(dtype):
    """Make sure that we don't immediately calculate dask arrays."""

    FETCH_COUNT = 0

    def get_plane(block_id):
        if isinstance(block_id, tuple):
            nonlocal FETCH_COUNT
            FETCH_COUNT += 1
        return np.random.rand(1, 1, 1, 10, 10)

    arr = da.map_blocks(
        get_plane,
        chunks=((1,) * 4, (1,) * 2, (1,) * 8, (10,), (10,)),
        dtype=dtype,
    )
    layer = layers.Image(arr)

    # the <= is because before dask-2021.12.0, the above code resulted in NO
    # fetches for uint8 data, and afterwards, results in a single fetch.
    # the single fetch is actually the more "expected" behavior.  And all we
    # are really trying to assert here is that we didn't fetch all the planes
    # in the first index... so we allow 0-1 fetches.
    assert FETCH_COUNT <= 1
    if dtype == 'uint8':
        assert tuple(layer.contrast_limits) == (0, 255)


def test_dask_array_creates_cache():
    """Test that dask arrays create cache but turns off fusion."""
    resize_dask_cache(1)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 1
    # by default we have no dask_cache and task fusion is active
    original = dask.config.get("optimization.fuse.active", None)

    def mock_set_view_slice():
        assert dask.config.get("optimization.fuse.active") is False

    layer = layers.Image(da.ones((100, 100)))
    layer._set_view_slice = mock_set_view_slice
    layer.set_view_slice()

    # adding a dask array will reate cache and turn off task fusion,
    # *but only* during slicing (see "mock_set_view_slice" above)
    assert _dask_utils._DASK_CACHE.cache.available_bytes > 100
    assert not _dask_utils._DASK_CACHE.active
    assert dask.config.get("optimization.fuse.active", None) == original

    # make sure we can resize the cache
    resize_dask_cache(10000)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 10000

    # This should only affect dask arrays, and not numpy data
    def mock_set_view_slice2():
        assert dask.config.get("optimization.fuse.active", None) == original

    layer2 = layers.Image(np.ones((100, 100)))
    layer2._set_view_slice = mock_set_view_slice2
    layer2.set_view_slice()


def test_list_of_dask_arrays_doesnt_create_cache():
    """Test that adding a list of dask array also creates a dask cache."""
    resize_dask_cache(1)  # in case other tests created it
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 1
    original = dask.config.get("optimization.fuse.active", None)
    _ = layers.Image([da.ones((100, 100)), da.ones((20, 20))])
    assert _dask_utils._DASK_CACHE.cache.available_bytes > 100
    assert not _dask_utils._DASK_CACHE.active
    assert dask.config.get("optimization.fuse.active", None) == original


@pytest.fixture
def delayed_dask_stack():
    """A 4D (20, 10, 10, 10) delayed dask array, simulates disk io."""
    # we will return a dict with a 'calls' variable that tracks call count
    output = {'calls': 0}

    # create a delayed version of function that simply generates np.arrays
    # but also counts when it has been called
    @dask.delayed
    def get_array():
        nonlocal output
        output['calls'] += 1
        return np.random.rand(10, 10, 10)

    # then make a mock "timelapse" of 3D stacks
    # see https://napari.org/tutorials/applications/dask.html for details
    _list = [get_array() for fn in range(20)]
    output['stack'] = da.stack(
        [da.from_delayed(i, shape=(10, 10, 10), dtype=float) for i in _list]
    )
    assert output['stack'].shape == (20, 10, 10, 10)
    return output


@pytest.mark.sync_only
def test_dask_global_optimized_slicing(delayed_dask_stack, monkeypatch):
    """Test that dask_configure reduces compute with dask stacks."""

    # add dask stack to the viewer, making sure to pass multiscale and clims
    v = ViewerModel()
    dask_stack = delayed_dask_stack['stack']
    layer = v.add_image(dask_stack)
    # the first and the middle stack will be loaded
    assert delayed_dask_stack['calls'] == 2

    with layer.dask_optimized_slicing() as (_, cache):
        assert cache.cache.available_bytes > 0
        assert cache.active
        # make sure the cache actually has been populated
        assert len(cache.cache.heap.heap) > 0

    assert not cache.active  # only active inside of the context

    # changing the Z plane should never incur calls
    # since the stack has already been loaded (& it is chunked as a 3D array)
    current_z = v.dims.point[1]
    for i in range(3):
        v.dims.set_point(1, current_z + i)
        assert delayed_dask_stack['calls'] == 2  # still just the first call

    # changing the timepoint will, of course, incur some compute calls
    initial_t = v.dims.point[0]
    v.dims.set_point(0, initial_t + 1)
    assert delayed_dask_stack['calls'] == 3
    v.dims.set_point(0, initial_t + 2)
    assert delayed_dask_stack['calls'] == 4

    # but going back to previous timepoints should not, since they are cached
    v.dims.set_point(0, initial_t + 1)
    v.dims.set_point(0, initial_t + 0)
    assert delayed_dask_stack['calls'] == 4
    # again, visiting a new point will increment the counter
    v.dims.set_point(0, initial_t + 3)
    assert delayed_dask_stack['calls'] == 5


@pytest.mark.sync_only
def test_dask_unoptimized_slicing(delayed_dask_stack, monkeypatch):
    """Prove that the dask_configure function works with a counterexample."""
    # we start with a cache...but then intentionally turn it off per-layer.
    resize_dask_cache(10000)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 10000

    # add dask stack to viewer.
    v = ViewerModel()
    dask_stack = delayed_dask_stack['stack']
    layer = v.add_image(dask_stack, cache=False)
    # the first and the middle stack will be loaded
    assert delayed_dask_stack['calls'] == 2

    with layer.dask_optimized_slicing() as (_, cache):
        assert cache is None

    # without optimized dask slicing, we get a new call to the get_array func
    # (which "re-reads" the full z stack) EVERY time we change the Z plane
    # even though we've already read this full timepoint.
    current_z = v.dims.point[1]
    for i in range(3):
        v.dims.set_point(1, current_z + i)
        assert delayed_dask_stack['calls'] == 2 + i  # 😞

    # of course we still incur calls when moving to a new timepoint...
    initial_t = v.dims.point[0]
    v.dims.set_point(0, initial_t + 1)
    v.dims.set_point(0, initial_t + 2)
    assert delayed_dask_stack['calls'] == 6

    # without the cache we ALSO incur calls when returning to previously loaded
    # timepoints 😭
    v.dims.set_point(0, initial_t + 1)
    v.dims.set_point(0, initial_t + 0)
    v.dims.set_point(0, initial_t + 3)
    # all told, we have ~2x as many calls as the optimized version above.
    # (should be exactly 9 calls, but for some reason, sometimes more on CI)
    assert delayed_dask_stack['calls'] >= 9


@pytest.mark.sync_only
def test_dask_local_unoptimized_slicing(delayed_dask_stack, monkeypatch):
    """Prove that the dask_configure function works with a counterexample."""
    # make sure we are not caching for this test, which also tests that we
    # can turn off caching
    resize_dask_cache(0)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 0

    monkeypatch.setattr(
        layers.base.base, 'configure_dask', lambda *_: nullcontext
    )

    # add dask stack to viewer.
    v = ViewerModel()
    dask_stack = delayed_dask_stack['stack']
    v.add_image(dask_stack, cache=False)
    # the first and the middle stack will be loaded
    assert delayed_dask_stack['calls'] == 2

    # without optimized dask slicing, we get a new call to the get_array func
    # (which "re-reads" the full z stack) EVERY time we change the Z plane
    # even though we've already read this full timepoint.
    for i in range(3):
        v.dims.set_point(1, i)
        assert delayed_dask_stack['calls'] == 2 + 1 + i  # 😞

    # of course we still incur calls when moving to a new timepoint...
    v.dims.set_point(0, 1)
    v.dims.set_point(0, 2)
    assert delayed_dask_stack['calls'] == 7

    # without the cache we ALSO incur calls when returning to previously loaded
    # timepoints 😭
    v.dims.set_point(0, 1)
    v.dims.set_point(0, 0)
    v.dims.set_point(0, 3)
    # all told, we have ~2x as many calls as the optimized version above.
    # (should be exactly 8 calls, but for some reason, sometimes less on CI)
    assert delayed_dask_stack['calls'] >= 10


@pytest.mark.sync_only
def test_dask_cache_resizing(delayed_dask_stack):
    """Test that we can spin up, resize, and spin down the cache."""

    # make sure we have a cache
    # big enough for 10+ (10, 10, 10) "timepoints"
    resize_dask_cache(100000)

    # add dask stack to the viewer, making sure to pass multiscale and clims

    v = ViewerModel()
    dask_stack = delayed_dask_stack['stack']

    v.add_image(dask_stack)
    assert _dask_utils._DASK_CACHE.cache.available_bytes > 0
    # make sure the cache actually has been populated
    assert len(_dask_utils._DASK_CACHE.cache.heap.heap) > 0

    # we can resize that cache back to 0 bytes
    resize_dask_cache(0)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 0

    # adding a 2nd stack should not adjust the cache size once created
    v.add_image(dask_stack)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 0
    # and the cache will remain empty regardless of what we do
    for i in range(3):
        v.dims.set_point(1, i)
    assert len(_dask_utils._DASK_CACHE.cache.heap.heap) == 0

    # but we can always spin it up again
    resize_dask_cache(1e4)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 1e4
    # and adding a new image doesn't change the size
    v.add_image(dask_stack)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 1e4
    # but the cache heap is getting populated again
    for i in range(3):
        v.dims.set_point(0, i)
    assert len(_dask_utils._DASK_CACHE.cache.heap.heap) > 0


def test_prevent_dask_cache(delayed_dask_stack):
    """Test that pre-emptively setting cache to zero keeps it off"""
    resize_dask_cache(0)

    v = ViewerModel()
    dask_stack = delayed_dask_stack['stack']
    # adding a new stack will not increase the cache size
    v.add_image(dask_stack)
    assert _dask_utils._DASK_CACHE.cache.available_bytes == 0
    # and the cache will not be populated
    for i in range(3):
        v.dims.set_point(0, i)
    assert len(_dask_utils._DASK_CACHE.cache.heap.heap) == 0



import os
from pathlib import Path
from typing import Optional, Tuple

from pydantic import Field

from ..utils._base import _DEFAULT_CONFIG_PATH
from ..utils.translations import trans
from ._appearance import AppearanceSettings
from ._application import ApplicationSettings
from ._base import EventedConfigFileSettings, _remove_empty_dicts
from ._experimental import ExperimentalSettings
from ._plugins import PluginsSettings
from ._shortcuts import ShortcutsSettings

_CFG_PATH = os.getenv('NAPARI_CONFIG', _DEFAULT_CONFIG_PATH)


class NapariSettings(EventedConfigFileSettings):
    """Schema for napari settings."""

    # 1. If you want to *change* the default value of a current option, you need to
    #    do a MINOR update in config version, e.g. from 3.0.0 to 3.1.0
    # 2. If you want to *remove* options that are no longer needed in the codebase,
    #    or if you want to *rename* options, then you need to do a MAJOR update in
    #    version, e.g. from 3.0.0 to 4.0.0
    # 3. You don't need to touch this value if you're just adding a new option
    schema_version: Tuple[int, int, int] = Field(
        (0, 3, 0),
        description=trans._("Napari settings schema version."),
        allow_mutation=False,
    )
    application: ApplicationSettings = Field(
        default_factory=ApplicationSettings,
        title=trans._("Application"),
        description=trans._("Main application settings."),
    )
    appearance: AppearanceSettings = Field(
        default_factory=AppearanceSettings,
        title=trans._("Appearance"),
        description=trans._("User interface appearance settings."),
    )
    plugins: PluginsSettings = Field(
        default_factory=PluginsSettings,
        title=trans._("Plugins"),
        description=trans._("Plugins settings."),
    )
    shortcuts: ShortcutsSettings = Field(
        default_factory=ShortcutsSettings,
        title=trans._("Shortcuts"),
        description=trans._("Shortcut settings."),
    )
    experimental: ExperimentalSettings = Field(
        default_factory=ExperimentalSettings,
        title=trans._("Experimental"),
        description=trans._("Experimental settings."),
    )

    # private attributes and ClassVars will not appear in the schema
    _config_path: Optional[Path] = Path(_CFG_PATH) if _CFG_PATH else None

    class Config:
        env_prefix = 'napari_'
        use_enum_values = False
        # all of these fields are evented models, so we don't want to break
        # connections by setting the top-level field itself
        # (you can still mutate attributes in the subfields)
        allow_mutation = False

    def __str__(self):
        out = 'NapariSettings (defaults excluded)\n' + 34 * '-' + '\n'
        data = self.dict(exclude_defaults=True)
        out += self._yaml_dump(_remove_empty_dicts(data))
        return out

    def __repr__(self):
        return str(self)


if __name__ == '__main__':
    import sys

    if len(sys.argv) > 2:
        dest = Path(sys.argv[2]).expanduser().absolute()
    else:
        dest = Path(__file__).parent / 'napari.schema.json'
    dest.write_text(NapariSettings.schema_json())



from collections import namedtuple

import numpy as np
import pytest

from napari._qt.layer_controls.qt_layer_controls_container import (
    create_qt_layer_controls,
    layer_to_controls,
)
from napari._qt.layer_controls.qt_shapes_controls import QtShapesControls
from napari.layers import Points, Shapes

LayerTypeWithData = namedtuple('LayerTypeWithData', ['type', 'data'])
_POINTS = LayerTypeWithData(type=Points, data=np.random.random((5, 2)))
_SHAPES = LayerTypeWithData(type=Shapes, data=np.random.random((10, 4, 2)))
_LINES_DATA = np.random.random((6, 2, 2))


def test_create_shape(qtbot):
    shapes = _SHAPES.type(_SHAPES.data)

    ctrl = create_qt_layer_controls(shapes)
    qtbot.addWidget(ctrl)

    assert isinstance(ctrl, QtShapesControls)


def test_unknown_raises(qtbot):
    class Test:
        """Unmatched class"""

    with pytest.raises(TypeError):
        create_qt_layer_controls(Test())


def test_inheritance(qtbot):
    class QtLinesControls(QtShapesControls):
        """Yes I'm the same"""

    class Lines(Shapes):
        """Here too"""

    lines = Lines(_LINES_DATA)
    layer_to_controls[Lines] = QtLinesControls
    ctrl = create_qt_layer_controls(lines)
    qtbot.addWidget(ctrl)
    assert isinstance(ctrl, QtLinesControls)


@pytest.mark.parametrize('layer_type_with_data', [_POINTS, _SHAPES])
def test_text_set_visible_updates_checkbox(qtbot, layer_type_with_data):
    text = {
        'text': 'test',
        'visible': True,
    }
    layer = layer_type_with_data.type(layer_type_with_data.data, text=text)
    ctrl = create_qt_layer_controls(layer)
    qtbot.addWidget(ctrl)
    assert ctrl.textDispCheckBox.isChecked()

    layer.text.visible = False

    assert not ctrl.textDispCheckBox.isChecked()


@pytest.mark.parametrize('layer_type_with_data', [_POINTS, _SHAPES])
def test_set_text_then_set_visible_updates_checkbox(
    qtbot, layer_type_with_data
):
    layer = layer_type_with_data.type(layer_type_with_data.data)
    ctrl = create_qt_layer_controls(layer)
    qtbot.addWidget(ctrl)
    layer.text = {
        'text': 'another_test',
        'visible': False,
    }
    assert not ctrl.textDispCheckBox.isChecked()

    layer.text.visible = True

    assert ctrl.textDispCheckBox.isChecked()



import threading
import time
import warnings
from concurrent.futures import Future
from unittest.mock import patch

import dask.array as da
import pytest
from qtpy.QtCore import Qt, QThread
from qtpy.QtWidgets import QPushButton

from napari._qt.dialogs.qt_notification import NapariQtNotification
from napari.utils.notifications import (
    ErrorNotification,
    Notification,
    NotificationSeverity,
    notification_manager,
)


def _threading_warn():
    thr = threading.Thread(target=_warn)
    thr.start()


def _warn():
    time.sleep(0.1)
    warnings.warn('warning!')


def _threading_raise():
    thr = threading.Thread(target=_raise)
    thr.start()


def _raise():
    time.sleep(0.1)
    raise ValueError("error!")


@pytest.fixture
def clean_current(monkeypatch, qtbot):
    from napari._qt.qt_main_window import _QtMainWindow

    def none_return(*_, **__):
        return None

    base_show = NapariQtNotification.show

    def store_widget(self, *args, **kwargs):
        qtbot.addWidget(self)
        base_show(self, *args, **kwargs)

    # monkeypatch.setattr(qt_notification.QPropertyAnimation, "start", none_return)
    monkeypatch.setattr(_QtMainWindow, "current", none_return)
    monkeypatch.setattr(NapariQtNotification, "show", store_widget)


@pytest.mark.parametrize(
    "raise_func,warn_func",
    [(_raise, _warn), (_threading_raise, _threading_warn)],
)
@pytest.mark.order(11)
def test_notification_manager_via_gui(
    qtbot, raise_func, warn_func, clean_current
):
    """
    Test that the notification_manager intercepts `sys.excepthook`` and
    `threading.excepthook`.
    """
    errButton = QPushButton()
    warnButton = QPushButton()
    errButton.clicked.connect(raise_func)
    warnButton.clicked.connect(warn_func)

    with notification_manager:
        for btt, expected_message in [
            (errButton, 'error!'),
            (warnButton, 'warning!'),
        ]:
            notification_manager.records = []
            qtbot.mouseClick(btt, Qt.LeftButton)
            qtbot.wait(500)
            assert len(notification_manager.records) == 1
            assert notification_manager.records[0].message == expected_message
            notification_manager.records = []


@patch('napari._qt.dialogs.qt_notification.QDialog.show')
def test_show_notification_from_thread(mock_show, monkeypatch, qtbot):
    from napari.settings import get_settings

    settings = get_settings()

    monkeypatch.setattr(
        settings.application,
        'gui_notification_level',
        NotificationSeverity.INFO,
    )

    class CustomThread(QThread):
        def run(self):
            notif = Notification(
                'hi',
                NotificationSeverity.INFO,
                actions=[('click', lambda x: None)],
            )
            res = NapariQtNotification.show_notification(notif)
            assert isinstance(res, Future)
            assert res.result() is None
            mock_show.assert_called_once()

    thread = CustomThread()
    with qtbot.waitSignal(thread.finished):
        thread.start()


@pytest.mark.parametrize('severity', NotificationSeverity.__members__)
@patch('napari._qt.dialogs.qt_notification.QDialog.show')
def test_notification_display(mock_show, severity, monkeypatch, qtbot):
    """Test that NapariQtNotification can present a Notification event.

    NOTE: in napari.utils._tests.test_notification_manager, we already test
    that the notification manager successfully overrides sys.excepthook,
    and warnings.showwarning... and that it emits an event which is an instance
    of napari.utils.notifications.Notification.

    in `get_app()`, we connect `notification_manager.notification_ready` to
    `NapariQtNotification.show_notification`, so all we have to test here is
    that show_notification is capable of receiving various event types.
    (we don't need to test that )
    """
    from napari.settings import get_settings

    settings = get_settings()

    monkeypatch.delenv('NAPARI_CATCH_ERRORS', raising=False)
    monkeypatch.setattr(
        settings.application,
        'gui_notification_level',
        NotificationSeverity.INFO,
    )
    notif = Notification('hi', severity, actions=[('click', lambda x: None)])
    NapariQtNotification.show_notification(notif)
    if NotificationSeverity(severity) >= NotificationSeverity.INFO:
        mock_show.assert_called_once()
    else:
        mock_show.assert_not_called()

    dialog = NapariQtNotification.from_notification(notif)
    assert not dialog.property('expanded')
    dialog.toggle_expansion()
    assert dialog.property('expanded')
    dialog.toggle_expansion()
    assert not dialog.property('expanded')
    dialog.close()
    dialog.deleteLater()


@patch('napari._qt.dialogs.qt_notification.QDialog.show')
def test_notification_error(mock_show, monkeypatch, clean_current):
    from napari.settings import get_settings

    settings = get_settings()

    monkeypatch.delenv('NAPARI_CATCH_ERRORS', raising=False)
    monkeypatch.setattr(
        settings.application,
        'gui_notification_level',
        NotificationSeverity.INFO,
    )
    try:
        raise ValueError('error!')
    except ValueError as e:
        notif = ErrorNotification(e)

    dialog = NapariQtNotification.from_notification(notif)
    bttn = dialog.row2_widget.findChild(QPushButton)
    assert bttn.text() == 'View Traceback'
    mock_show.assert_not_called()
    bttn.click()
    mock_show.assert_called_once()


@pytest.mark.sync_only
def test_notifications_error_with_threading(make_napari_viewer):
    """Test notifications of `threading` threads, using a dask example."""
    random_image = da.random.random((50, 50))
    with notification_manager:
        viewer = make_napari_viewer()
        viewer.add_image(random_image)
        result = da.divide(random_image, da.zeros((50, 50)))
        viewer.add_image(result)
        assert len(notification_manager.records) >= 1
        notification_manager.records = []



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple

from ...utils._dtype import normalize_dtype
from ...utils.translations import trans
from ._context_keys import ContextKey, ContextNamespace

if TYPE_CHECKING:
    from numpy.typing import DTypeLike

    from ...layers import Layer
    from ...utils.events import Selection

    LayerSel = Selection[Layer]


def _len(s: LayerSel) -> int:
    return len(s)


def _all_linked(s: LayerSel) -> bool:
    from ...layers.utils._link_layers import layer_is_linked

    return bool(s and all(layer_is_linked(x) for x in s))


def _n_unselected_links(s: LayerSel) -> int:
    from ...layers.utils._link_layers import get_linked_layers

    return len(get_linked_layers(*s) - s)


def _is_rgb(s: LayerSel) -> bool:
    return getattr(s.active, "rgb", False)


def _only_img(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "image" for x in s))


def _only_labels(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "labels" for x in s))


def _only_shapes(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "shapes" for x in s))


def _active_type(s: LayerSel) -> Optional[str]:
    return s.active._type_string if s.active else None


def _active_ndim(s: LayerSel) -> Optional[int]:
    return getattr(s.active.data, "ndim", None) if s.active else None


def _active_shape(s: LayerSel) -> Optional[Tuple[int, ...]]:
    return getattr(s.active.data, "shape", None) if s.active else None


def _same_shape(s: LayerSel) -> bool:
    return len({getattr(x.data, "shape", ()) for x in s}) == 1


def _active_dtype(s: LayerSel) -> DTypeLike:
    dtype = None
    if s.active:
        try:
            dtype = normalize_dtype(s.active.data.dtype).__name__
        except AttributeError:
            pass
    return dtype


class LayerListContextKeys(ContextNamespace['LayerSel']):
    """These are the available context keys relating to a LayerList.

    along with default value, a description, and a function to retrieve the
    current value from layers.selection
    """

    layers_selection_count = ContextKey(
        0,
        trans._("Number of layers currently selected"),
        _len,
    )
    all_layers_linked = ContextKey(
        False,
        trans._("True when all selected layers are linked."),
        _all_linked,
    )
    unselected_linked_layers = ContextKey(
        0,
        trans._("Number of unselected layers linked to selected layer(s)"),
        _n_unselected_links,
    )
    active_layer_is_rgb = ContextKey(
        False,
        trans._("True when the active layer is RGB"),
        _is_rgb,
    )
    active_layer_type = ContextKey['LayerSel', Optional[str]](
        None,
        trans._(
            "Lowercase name of active layer type, or None of none active."
        ),
        _active_type,
    )
    # TODO: try to reduce these `only_x_selected` to a single set of strings
    # or something... however, this would require that our context expressions
    # support Sets, tuples, lists, etc...  which they currently do not.
    only_images_selected = ContextKey(
        False,
        trans._(
            "True when there is at least one selected layer and all selected layers are images"
        ),
        _only_img,
    )
    only_labels_selected = ContextKey(
        False,
        trans._(
            "True when there is at least one selected layer and all selected layers are labels"
        ),
        _only_labels,
    )
    only_shapes_selected = ContextKey(
        False,
        trans._(
            "True when there is at least one selected layer and all selected layers are shapes"
        ),
        _only_shapes,
    )
    active_layer_ndim = ContextKey['LayerSel', Optional[int]](
        None,
        trans._(
            "Number of dimensions in the active layer, or `None` if nothing is active"
        ),
        _active_ndim,
    )
    active_layer_shape = ContextKey['LayerSel', Optional[Tuple[int, ...]]](
        (),
        trans._("Shape of the active layer, or `None` if nothing is active."),
        _active_shape,
    )
    active_layer_dtype = ContextKey(
        None,
        trans._("Dtype of the active layer, or `None` if nothing is active."),
        _active_dtype,
    )
    all_layers_same_shape = ContextKey(
        False,
        trans._("True when all selected layers have the same shape"),
        _same_shape,
    )



import numpy as np

from napari._vispy.layers.shapes import VispyShapesLayer
from napari.layers import Shapes


def test_remove_selected_with_derived_text():
    """See https://github.com/napari/napari/issues/3504"""
    shapes = np.random.rand(3, 4, 2)
    properties = {'class': np.array(['A', 'B', 'C'])}
    layer = Shapes(shapes, properties=properties, text='class')
    vispy_layer = VispyShapesLayer(layer)
    text_node = vispy_layer._get_text_node()
    np.testing.assert_array_equal(text_node.text, ['A', 'B', 'C'])

    layer.selected_data = {1}
    layer.remove_selected()

    np.testing.assert_array_equal(text_node.text, ['A', 'C'])


def test_change_text_updates_node_string():
    shapes = np.random.rand(3, 4, 2)
    properties = {
        'class': np.array(['A', 'B', 'C']),
        'name': np.array(['D', 'E', 'F']),
    }
    layer = Shapes(shapes, properties=properties, text='class')
    vispy_layer = VispyShapesLayer(layer)
    text_node = vispy_layer._get_text_node()
    np.testing.assert_array_equal(text_node.text, properties['class'])

    layer.text = 'name'

    np.testing.assert_array_equal(text_node.text, properties['name'])


def test_change_text_color_updates_node_color():
    shapes = np.random.rand(3, 4, 2)
    properties = {'class': np.array(['A', 'B', 'C'])}
    text = {'text': 'class', 'color': [1, 0, 0]}
    layer = Shapes(shapes, properties=properties, text=text)
    vispy_layer = VispyShapesLayer(layer)
    text_node = vispy_layer._get_text_node()
    np.testing.assert_array_equal(text_node.color.rgb, [[1, 0, 0]])

    layer.text.color = [0, 0, 1]

    np.testing.assert_array_equal(text_node.color.rgb, [[0, 0, 1]])


def test_change_properties_updates_node_strings():
    shapes = np.random.rand(3, 4, 2)
    properties = {'class': np.array(['A', 'B', 'C'])}
    layer = Shapes(shapes, properties=properties, text='class')
    vispy_layer = VispyShapesLayer(layer)
    text_node = vispy_layer._get_text_node()
    np.testing.assert_array_equal(text_node.text, ['A', 'B', 'C'])

    layer.properties = {'class': np.array(['D', 'E', 'F'])}

    np.testing.assert_array_equal(text_node.text, ['D', 'E', 'F'])


def test_update_property_value_then_refresh_text_updates_node_strings():
    shapes = np.random.rand(3, 4, 2)
    properties = {'class': np.array(['A', 'B', 'C'])}
    layer = Shapes(shapes, properties=properties, text='class')
    vispy_layer = VispyShapesLayer(layer)
    text_node = vispy_layer._get_text_node()
    np.testing.assert_array_equal(text_node.text, ['A', 'B', 'C'])

    layer.properties['class'][1] = 'D'
    layer.refresh_text()

    np.testing.assert_array_equal(text_node.text, ['A', 'D', 'C'])



import os
from unittest.mock import Mock

import pytest

from napari import Viewer
from napari._qt.qt_event_loop import _ipython_has_eventloop, run, set_app_id


@pytest.mark.skipif(os.name != "Windows", reason="Windows specific")
def test_windows_grouping_overwrite(make_napari_viewer):
    import ctypes

    def get_app_id():
        mem = ctypes.POINTER(ctypes.c_wchar)()
        ctypes.windll.shell32.GetCurrentProcessExplicitAppUserModelID(
            ctypes.byref(mem)
        )
        res = ctypes.wstring_at(mem)
        ctypes.windll.Ole32.CoTaskMemFree(mem)
        return res

    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("test_text")

    assert get_app_id() == "test_text"
    set_app_id("custom_string")
    assert get_app_id() == "custom_string"
    set_app_id("")
    assert get_app_id() == ""


def test_run_outside_ipython(qapp, monkeypatch):
    """Test that we don't incorrectly give ipython the event loop."""
    assert not _ipython_has_eventloop()
    v1 = Viewer(show=False)
    assert not _ipython_has_eventloop()
    v2 = Viewer(show=False)
    assert not _ipython_has_eventloop()

    with monkeypatch.context() as m:
        mock_exec = Mock()
        m.setattr(qapp, 'exec_', mock_exec)
        run()
        mock_exec.assert_called_once()

    v1.close()
    v2.close()



from vispy.visuals.filters import Filter


class ClampSizeFilter(Filter):
    """
    Clamp the size of the points in canvas pixels.
    """

    VERT_SHADER = """
        void clamp_size() {
            if ($active == 1) {
                gl_PointSize = clamp(gl_PointSize, $min, $max);
            }
        }
    """

    def __init__(self, min_size=0, max_size=10000, active=True):
        super().__init__(vcode=self.VERT_SHADER, vpos=10000)
        self.min_size = min_size
        self.max_size = max_size
        self.active = active

    @property
    def min_size(self):
        return self._min_size

    @min_size.setter
    def min_size(self, value):
        self._min_size = float(value)
        self.vshader['min'] = self._min_size

    @property
    def max_size(self):
        return self._max_size

    @max_size.setter
    def max_size(self, value):
        self._max_size = float(value)
        self.vshader['max'] = self._max_size

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, value):
        self._active = bool(value)
        self.vshader['active'] = self._active



from unittest.mock import patch

import pytest
from napari_plugin_engine import napari_hook_implementation

from napari._qt import Window
from napari.utils.theme import Theme, get_theme


@patch.object(Window, "_remove_theme")
@patch.object(Window, "_add_theme")
def test_provide_theme_hook_registered_correctly(
    mock_add_theme,
    mock_remove_theme,
    make_napari_viewer,
    napari_plugin_manager,
):
    with pytest.warns(FutureWarning):
        dark = get_theme("dark")
        dark["name"] = "dark-test-2"

    class TestPlugin:
        @napari_hook_implementation
        def napari_experimental_provide_theme():
            return {"dark-test-2": dark}

    # create instance of viewer to make sure
    # registration and unregistration methods are called
    viewer = make_napari_viewer()

    # register theme
    napari_plugin_manager.register(TestPlugin)
    reg = napari_plugin_manager._theme_data["TestPlugin"]
    assert isinstance(reg["dark-test-2"], Theme)

    viewer.theme = "dark-test-2"

    # triggered when theme was added
    mock_add_theme.assert_called()
    mock_remove_theme.assert_not_called()

    # now, lets unregister the theme
    with pytest.warns(UserWarning, match="The current theme "):
        napari_plugin_manager.unregister("TestPlugin")
    mock_remove_theme.assert_called()



from functools import partial

from ..utils.translations import trans

SKIMAGE_DATA = [
    ('astronaut', trans._('Astronaut (RGB)')),
    ('binary_blobs', trans._('Binary Blobs')),
    ('binary_blobs_3D', trans._('Binary Blobs (3D)')),
    ('brain', trans._('Brain (3D)')),
    ('brick', trans._('Brick')),
    ('camera', trans._('Camera')),
    ('cat', trans._('Cat (RGB)')),
    ('cell', trans._('Cell')),
    ('cells3d', trans._('Cells (3D+2Ch)')),
    ('checkerboard', trans._('Checkerboard')),
    ('clock', trans._('Clock')),
    ('coffee', trans._('Coffee (RGB)')),
    ('coins', trans._('Coins')),
    ('colorwheel', trans._('Colorwheel (RGB)')),
    ('eagle', trans._('Eagle')),
    ('grass', trans._('Grass')),
    ('gravel', trans._('Gravel')),
    ('horse', trans._('Horse')),
    ('hubble_deep_field', trans._('Hubble Deep Field (RGB)')),
    ('human_mitosis', trans._('Human Mitosis')),
    ('immunohistochemistry', trans._('Immunohistochemistry (RGB)')),
    ('kidney', trans._('Kidney (3D+3Ch)')),
    ('lfw_subset', trans._('Labeled Faces in the Wild')),
    ('lily', trans._('Lily (4Ch)')),
    ('microaneurysms', trans._('Microaneurysms')),
    ('moon', trans._('Moon')),
    ('page', trans._('Page')),
    ('retina', trans._('Retina (RGB)')),
    ('rocket', trans._('Rocket (RGB)')),
    ('shepp_logan_phantom', trans._('Shepp Logan Phantom')),
    ('skin', trans._('Skin (RGB)')),
    ('text', trans._('Text')),
]


def _load_skimage_data(name, **kwargs):
    import skimage.data

    if name == 'cells3d':
        return [
            (
                skimage.data.cells3d(),
                {
                    'channel_axis': 1,
                    'name': ['membrane', 'nuclei'],
                    'contrast_limits': [(1110, 23855), (1600, 50000)],
                },
            )
        ]
    elif name == 'kidney':
        return [
            (
                skimage.data.kidney(),
                {
                    'channel_axis': -1,
                    'name': ['nuclei', 'WGA', 'actin'],
                    'colormap': ['blue', 'green', 'red'],
                },
            )
        ]
    elif name == 'lily':
        return [
            (
                skimage.data.lily(),
                {
                    'channel_axis': -1,
                    'name': ['lily-R', 'lily-G', 'lily-W', 'lily-B'],
                    'colormap': ['red', 'green', 'gray', 'blue'],
                },
            )
        ]
    elif name == 'binary_blobs_3D':
        kwargs['n_dim'] = 3
        kwargs.setdefault('length', 128)
        kwargs.setdefault('volume_fraction', 0.25)
        name = 'binary_blobs'

    return [(getattr(skimage.data, name)(**kwargs), {'name': name})]


_DATA = {
    key: {'data': partial(_load_skimage_data, key), 'display_name': dname}
    for (key, dname) in SKIMAGE_DATA
}


globals().update({k: v['data'] for k, v in _DATA.items()})



import json
from enum import EnumMeta
from typing import TYPE_CHECKING, Tuple, cast

from pydantic.main import BaseModel
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtWidgets import (
    QDialog,
    QHBoxLayout,
    QListWidget,
    QMessageBox,
    QPushButton,
    QStackedWidget,
    QVBoxLayout,
)

from ...utils.translations import trans

if TYPE_CHECKING:
    from pydantic.fields import ModelField
    from qtpy.QtGui import QCloseEvent, QKeyEvent


class PreferencesDialog(QDialog):
    """Preferences Dialog for Napari user settings."""

    ui_schema = {
        "call_order": {"ui:widget": "plugins"},
        "highlight_thickness": {"ui:widget": "highlight"},
        "shortcuts": {"ui:widget": "shortcuts"},
        "extension2reader": {"ui:widget": "extension2reader"},
    }

    resized = Signal(QSize)

    def __init__(self, parent=None):
        from ...settings import get_settings

        super().__init__(parent)
        self.setWindowTitle(trans._("Preferences"))

        self._settings = get_settings()
        self._stack = QStackedWidget(self)
        self._list = QListWidget(self)
        self._list.setObjectName("Preferences")
        self._list.currentRowChanged.connect(self._stack.setCurrentIndex)

        # Set up buttons
        self._button_cancel = QPushButton(trans._("Cancel"))
        self._button_cancel.clicked.connect(self.reject)
        self._button_ok = QPushButton(trans._("OK"))
        self._button_ok.clicked.connect(self.accept)
        self._button_ok.setDefault(True)
        self._button_restore = QPushButton(trans._("Restore defaults"))
        self._button_restore.clicked.connect(self._restore_default_dialog)

        # Layout
        left_layout = QVBoxLayout()
        left_layout.addWidget(self._list)
        left_layout.addStretch()
        left_layout.addWidget(self._button_restore)
        left_layout.addWidget(self._button_cancel)
        left_layout.addWidget(self._button_ok)

        self.setLayout(QHBoxLayout())
        self.layout().addLayout(left_layout, 1)
        self.layout().addWidget(self._stack, 3)

        # Build dialog from settings
        self._rebuild_dialog()

    def keyPressEvent(self, e: 'QKeyEvent'):
        if e.key() == Qt.Key_Escape:
            # escape key should just close the window
            # which implies "accept"
            e.accept()
            self.accept()
            return
        super().keyPressEvent(e)

    def resizeEvent(self, event):
        """Override to emit signal."""
        self.resized.emit(event.size())
        super().resizeEvent(event)

    def _rebuild_dialog(self):
        """Removes settings not to be exposed to user and creates dialog pages."""
        # FIXME: this dialog should not need to know about the plugin manager
        from ...plugins import plugin_manager

        self._starting_pm_order = plugin_manager.call_order()
        self._starting_values = self._settings.dict(exclude={'schema_version'})

        self._list.clear()
        while self._stack.count():
            self._stack.removeWidget(self._stack.currentWidget())

        for field in self._settings.__fields__.values():
            if isinstance(field.type_, type) and issubclass(
                field.type_, BaseModel
            ):
                self._add_page(field)

        self._list.setCurrentRow(0)

    def _add_page(self, field: 'ModelField'):
        """Builds the preferences widget using the json schema builder.

        Parameters
        ----------
        field : ModelField
            subfield for which to create a page.
        """
        from ..._vendor.qt_json_builder.qt_jsonschema_form import WidgetBuilder

        schema, values = self._get_page_dict(field)
        name = field.field_info.title or field.name

        form = WidgetBuilder().create_form(schema, self.ui_schema)
        # set state values for widget
        form.widget.state = values
        # make settings follow state of the form widget
        form.widget.on_changed.connect(
            lambda d: getattr(self._settings, name.lower()).update(d)
        )

        # need to disable async if octree is enabled.
        # TODO: this shouldn't live here... if there is a coupling/dependency
        # between these settings, it should be declared in the settings schema
        if (
            name.lower() == 'experimental'
            and values['octree']
            and self._settings.env_settings()
            .get('experimental', {})
            .get('async_')
            not in (None, '0')
        ):
            form_layout = form.widget.layout()
            for i in range(form_layout.count()):
                wdg = form_layout.itemAt(i, form_layout.FieldRole).widget()
                if getattr(wdg, '_name') == 'async_':
                    wdg.opacity.setOpacity(0.3)
                    wdg.setDisabled(True)
                    break

        self._list.addItem(field.field_info.title or field.name)
        self._stack.addWidget(form)

    def _get_page_dict(self, field: 'ModelField') -> Tuple[dict, dict, dict]:
        """Provides the schema, set of values for each setting, and the
        properties for each setting."""
        ftype = cast('BaseModel', field.type_)
        schema = json.loads(ftype.schema_json())

        # find enums:
        for name, subfield in ftype.__fields__.items():
            if isinstance(subfield.type_, EnumMeta):
                enums = [s.value for s in subfield.type_]  # type: ignore
                schema["properties"][name]["enum"] = enums
                schema["properties"][name]["type"] = "string"

        # Need to remove certain properties that will not be displayed on the GUI
        setting = getattr(self._settings, field.name)
        with setting.enums_as_values():
            values = setting.dict()
        napari_config = getattr(setting, "NapariConfig", None)
        if hasattr(napari_config, 'preferences_exclude'):
            for val in napari_config.preferences_exclude:
                schema['properties'].pop(val, None)
                values.pop(val, None)

        return schema, values

    def _restore_default_dialog(self):
        """Launches dialog to confirm restore settings choice."""
        response = QMessageBox.question(
            self,
            trans._("Restore Settings"),
            trans._("Are you sure you want to restore default settings?"),
            QMessageBox.RestoreDefaults | QMessageBox.Cancel,
            QMessageBox.RestoreDefaults,
        )
        if response == QMessageBox.RestoreDefaults:
            self._settings.reset()
            self._rebuild_dialog()  # TODO: do we need this?

    def _restart_required_dialog(self):
        """Displays the dialog informing user a restart is required."""
        QMessageBox.information(
            self,
            trans._("Restart required"),
            trans._(
                "A restart is required for some new settings to have an effect."
            ),
        )

    def closeEvent(self, event: 'QCloseEvent') -> None:
        event.accept()
        self.accept()

    def reject(self):
        """Restores the settings in place when dialog was launched."""
        self._settings.update(self._starting_values)

        # FIXME: this dialog should not need to know about the plugin manager
        if self._starting_pm_order:
            from ...plugins import plugin_manager

            plugin_manager.set_call_order(self._starting_pm_order)
        super().reject()



import numpy as np

from napari.components.interaction_box import InteractionBox
from napari.utils.transforms import Affine


def test_creation():
    """Test creating interaction box object"""
    interaction_box = InteractionBox()
    assert interaction_box is not None


def test_box_from_points():
    """Test whether setting points creates a axis-aligned containing box"""
    interaction_box = InteractionBox()
    interaction_box.points = np.array([[1, 0], [3, 2], [-1, 1]])
    resulting_box = np.array(
        [
            [-1.0, 0.0],
            [1.0, 0.0],
            [3.0, 0.0],
            [3.0, 1.0],
            [3.0, 2.0],
            [1.0, 2.0],
            [-1.0, 2.0],
            [-1.0, 1.0],
            [1.0, 1.0],
        ]
    )
    np.testing.assert_equal(interaction_box._box, resulting_box)


def test_transform():
    """Tests whether setting a transform changes the box adequatly"""
    interaction_box = InteractionBox()
    interaction_box.points = np.array([[1, 0], [3, 2], [-1, 1]])
    resulting_box = np.array(
        [
            [-1.0, 0.0],
            [1.0, 0.0],
            [3.0, 0.0],
            [3.0, 1.0],
            [3.0, 2.0],
            [1.0, 2.0],
            [-1.0, 2.0],
            [-1.0, 1.0],
            [1.0, 1.0],
        ]
    )
    interaction_box.transform = Affine(rotate=45)
    resulting_box = Affine(rotate=45)(resulting_box)
    np.testing.assert_equal(interaction_box._box, resulting_box)



"""ChunkCache stores loaded chunks.
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Optional

from ...._vendor.experimental.cachetools import LRUCache

if TYPE_CHECKING:
    from ....types import ArrayLike
    from ._request import ChunkRequest

    # A ChunkRequest is just a dict of the arrays we need to load. We allow
    # loading multiple arrays in one request so the caller does not have to
    # deal with partial loads, where it has received some arrays but it cannot
    # use them until other arrays have finished loading.
    #
    # The caller is free to use whatever names it wants to organize the arrays,
    # for example "image" and "thumbnail", or spatially neighboring tiles like
    # "tile.1.1", "tile1.2", "tile2.1", "tile2.2".
    ChunkArrays = Dict[str, ArrayLike]


LOGGER = logging.getLogger("napari.loader.cache")

# ChunkCache size as a fraction of total RAM. Keep it small for now until
# we figure out how ChunkCache will work with the Dask cache, and do
# a lot more testing.
CACHE_MEM_FRACTION = 0.1


def _get_cache_size_bytes(mem_fraction: float) -> int:
    """Return the max number of bytes the cache should use.

    Parameters
    ----------
    mem_fraction : float
        The cache should use this fraction of RAM, for example 0.5.

    Returns
    -------
    int
        The max number of bytes the cache should use.
    """
    import psutil

    return psutil.virtual_memory().total * mem_fraction


def _getsizeof_chunks(chunks: ChunkArrays) -> int:
    """This tells the LRUCache know how big our chunks are.

    Parameters
    ----------
    chunks : ChunkArrays
        The arrays stored in one cache entry.

    Returns
    -------
    int
        How many bytes the arrays take up in memory.
    """
    return sum(array.nbytes for array in chunks.values())


class ChunkCache:
    """Cache of previously loaded chunks.

    We use a cachetools LRUCache to implement a least recently used cache
    that will grow in memory usage up to some limit. Then it will free the
    least recently used entries so total usage does not exceed that limit.

    TODO_OCTREE:

    1) For dynamically computed data the cache should be disabled. So
       should the default be off? Or can we detect dynamic computations?

    2) Can we use the Dask cache instead of our own? The problem with having
       two caches is how to manage their sizes? They can't both be 0.5 * RAM
       for example!

    Attributes
    ----------
    chunks : LRUCache
        The cache of chunks.
    enabled : bool
        True if the cache is enabled.
    """

    def __init__(self):
        nbytes = _get_cache_size_bytes(CACHE_MEM_FRACTION)
        self.chunks = LRUCache(maxsize=nbytes, getsizeof=_getsizeof_chunks)
        self.enabled = True

    def add_chunks(self, request: ChunkRequest) -> None:
        """Add the chunks in this request to the cache.

        Parameters
        ----------
        request : ChunkRequest
            Add the data in this request to the cache.
        """
        if not self.enabled:
            LOGGER.debug("ChunkCache.add_chunk: cache is disabled")
            return
        LOGGER.debug("add_chunk: %s", request.location)
        self.chunks[request.location] = request.chunks

    def get_chunks(self, request: ChunkRequest) -> Optional[ChunkArrays]:
        """Return the cached data for this request if it was cached.

        Parameters
        ----------
        request : ChunkRequest
            Look for cached data for this request.

        Returns
        -------
        Optional[ChunkArrays]
            The cached data or None of it was not found in the cache.
        """
        if not self.enabled:
            LOGGER.info("ChunkCache.get_chunk: disabled")
            return None

        data = self.chunks.get(request.location)
        LOGGER.info(
            "get_chunk: %s %s",
            request.location,
            "found" if data is not None else "not found",
        )
        return data



import numpy as np


class Mesh:
    """Contains meshses of shapes that will ultimately get rendered.

    Parameters
    ----------
    ndisplay : int
        Number of displayed dimensions.

    Attributes
    ----------
    ndisplay : int
        Number of displayed dimensions.
    vertices : np.ndarray
         Qx2 array of vertices of all triangles for shapes including edges and
         faces
    vertices_centers : np.ndarray
         Qx2 array of centers of vertices of triangles for shapes. For vertices
         corresponding to faces these are the same as the actual vertices. For
         vertices corresponding to edges these values should be added to a
         scaled `vertices_offsets` to get the actual vertex positions.
         The scaling corresponds to the width of the edge
    vertices_offsets : np.ndarray
         Qx2 array of offsets of vertices of triangles for shapes. For vertices
         corresponding to faces these are 0. For vertices corresponding to
         edges these values should be scaled and added to the
         `vertices_centers` to get the actual vertex positions.
         The scaling corresponds to the width of the edge
    vertices_index : np.ndarray
         Qx2 array of the index (0, ..., N-1) of each shape that each vertex
         corresponds and the mesh type (0, 1) for face or edge.
    triangles : np.ndarray
        Px3 array of vertex indices that form the mesh triangles
    triangles_index : np.ndarray
        Px2 array of  the index (0, ..., N-1) of each shape that each triangle
        corresponds and the mesh type (0, 1) for face or edge.
    triangles_colors : np.ndarray
        Px4 array of the rgba color of each triangle
    triangles_z_order : np.ndarray
        Length P array of the z order of each triangle. Must be a permutation
        of (0, ..., P-1)

    Notes
    -----
    _types : list
        Length two list of the different mesh types corresponding to faces and
        edges
    """

    _types = ['face', 'edge']

    def __init__(self, ndisplay=2):

        self._ndisplay = ndisplay
        self.clear()

    def clear(self):
        """Resets mesh data"""
        self.vertices = np.empty((0, self.ndisplay))
        self.vertices_centers = np.empty((0, self.ndisplay))
        self.vertices_offsets = np.empty((0, self.ndisplay))
        self.vertices_index = np.empty((0, 2), dtype=int)
        self.triangles = np.empty((0, 3), dtype=np.uint32)
        self.triangles_index = np.empty((0, 2), dtype=int)
        self.triangles_colors = np.empty((0, 4))
        self.triangles_z_order = np.empty((0), dtype=int)

        self.displayed_triangles = np.empty((0, 3), dtype=np.uint32)
        self.displayed_triangles_index = np.empty((0, 2), dtype=int)
        self.displayed_triangles_colors = np.empty((0, 4))

    @property
    def ndisplay(self):
        """int: Number of displayed dimensions."""
        return self._ndisplay

    @ndisplay.setter
    def ndisplay(self, ndisplay):
        if self.ndisplay == ndisplay:
            return

        self._ndisplay = ndisplay
        self.clear()



# syntax_style for the console must be one of the supported styles from
# pygments - see here for examples https://help.farbox.com/pygments.html
import re
import warnings
from ast import literal_eval
from typing import Union

from pydantic import validator
from pydantic.color import Color

try:
    from qtpy import QT_VERSION

    major, minor, *rest = QT_VERSION.split('.')
    use_gradients = (int(major) >= 5) and (int(minor) >= 12)
except Exception:
    use_gradients = False

from .._vendor import darkdetect
from ..utils.translations import trans
from .events import EventedModel
from .events.containers._evented_dict import EventedDict


class Theme(EventedModel):
    """Theme model.

    Attributes
    ----------
    name : str
        Name of the virtual folder where icons will be saved to.
    syntax_style : str
        Name of the console style.
        See for more details: https://pygments.org/docs/styles/
    canvas : Color
        Background color of the canvas.
    background : Color
        Color of the application background.
    foreground : Color
        Color to contrast with the background.
    primary : Color
        Color used to make part of a widget more visible.
    secondary : Color
        Alternative color used to make part of a widget more visible.
    highlight : Color
        Color used to highlight visual element.
    text : Color
        Color used to display text.
    warning : Color
        Color used to indicate something is wrong.
    current : Color
        Color used to highlight Qt widget.
    """

    name: str
    syntax_style: str
    canvas: Color
    console: Color
    background: Color
    foreground: Color
    primary: Color
    secondary: Color
    highlight: Color
    text: Color
    icon: Color
    warning: Color
    current: Color

    @validator("syntax_style", pre=True)
    def _ensure_syntax_style(value: str) -> str:
        from pygments.styles import STYLE_MAP

        assert value in STYLE_MAP, trans._(
            "Incorrect `syntax_style` value provided. Please use one of the following: {syntax_style}",
            deferred=True,
            syntax_style=f" {', '.join(STYLE_MAP)}",
        )
        return value


gradient_pattern = re.compile(r'([vh])gradient\((.+)\)')
darken_pattern = re.compile(r'{{\s?darken\((\w+),?\s?([-\d]+)?\)\s?}}')
lighten_pattern = re.compile(r'{{\s?lighten\((\w+),?\s?([-\d]+)?\)\s?}}')
opacity_pattern = re.compile(r'{{\s?opacity\((\w+),?\s?([-\d]+)?\)\s?}}')


def darken(color: Union[str, Color], percentage=10):
    if isinstance(color, str) and color.startswith('rgb('):
        color = literal_eval(color.lstrip('rgb(').rstrip(')'))
    else:
        color = color.as_rgb_tuple()
    ratio = 1 - float(percentage) / 100
    red, green, blue = color
    red = min(max(int(red * ratio), 0), 255)
    green = min(max(int(green * ratio), 0), 255)
    blue = min(max(int(blue * ratio), 0), 255)
    return f'rgb({red}, {green}, {blue})'


def lighten(color: Union[str, Color], percentage=10):
    if isinstance(color, str) and color.startswith('rgb('):
        color = literal_eval(color.lstrip('rgb(').rstrip(')'))
    else:
        color = color.as_rgb_tuple()
    ratio = float(percentage) / 100
    red, green, blue = color
    red = min(max(int(red + (255 - red) * ratio), 0), 255)
    green = min(max(int(green + (255 - green) * ratio), 0), 255)
    blue = min(max(int(blue + (255 - blue) * ratio), 0), 255)
    return f'rgb({red}, {green}, {blue})'


def opacity(color: Union[str, Color], value=255):
    if isinstance(color, str) and color.startswith('rgb('):
        color = literal_eval(color.lstrip('rgb(').rstrip(')'))
    else:
        color = color.as_rgb_tuple()
    red, green, blue = color
    return f'rgba({red}, {green}, {blue}, {max(min(int(value), 255), 0)})'


def gradient(stops, horizontal=True):
    if not use_gradients:
        return stops[-1]

    if horizontal:
        grad = 'qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, '
    else:
        grad = 'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1, '

    _stops = [f'stop: {n} {stop}' for n, stop in enumerate(stops)]
    grad += ", ".join(_stops) + ")"

    return grad


def template(css: str, **theme):
    def darken_match(matchobj):
        color, percentage = matchobj.groups()
        return darken(theme[color], percentage)

    def lighten_match(matchobj):
        color, percentage = matchobj.groups()
        return lighten(theme[color], percentage)

    def opacity_match(matchobj):
        color, percentage = matchobj.groups()
        return opacity(theme[color], percentage)

    def gradient_match(matchobj):
        horizontal = matchobj.groups()[1] == 'h'
        stops = [i.strip() for i in matchobj.groups()[1].split('-')]
        return gradient(stops, horizontal)

    for k, v in theme.items():
        css = gradient_pattern.sub(gradient_match, css)
        css = darken_pattern.sub(darken_match, css)
        css = lighten_pattern.sub(lighten_match, css)
        css = opacity_pattern.sub(opacity_match, css)
        if isinstance(v, Color):
            v = v.as_rgb()
        css = css.replace('{{ %s }}' % k, v)
    return css


def get_system_theme():
    """Return the system default theme, either 'dark', or 'light'."""
    try:
        name = darkdetect.theme().lower()
    except Exception:
        name = "dark"

    return name


def get_theme(name, as_dict=None):
    """Get a copy of theme based on it's name.

    If you get a copy of the theme, changes to the theme model will not be
    reflected in the UI unless you replace or add the modified theme to
    the `_themes` container.

    Parameters
    ----------
    name : str
        Name of requested theme.
    as_dict : bool
        Flag to indicate that the old-style dictionary
        should be returned. This will emit deprecation warning.

    Returns
    -------
    theme: dict of str: str
        Theme mapping elements to colors. A copy is created
        so that manipulating this theme can be done without
        side effects.
    """
    if name == "system":
        name = get_system_theme()

    if name not in _themes:
        raise ValueError(
            trans._(
                "Unrecognized theme {name}. Available themes are {themes}",
                deferred=True,
                name=name,
                themes=available_themes(),
            )
        )
    theme = _themes[name]
    _theme = theme.copy()
    if as_dict is None:
        warnings.warn(
            trans._(
                "Themes were changed to use evented model with Pydantic's color type rather than the `rgb(x, y, z)`. The `as_dict=True` option will be changed to `as_dict=False` in 0.4.15",
                deferred=True,
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        as_dict = True
    if as_dict:
        _theme = _theme.dict()
        _theme = {
            k: v if not isinstance(v, Color) else v.as_rgb()
            for (k, v) in _theme.items()
        }
        return _theme
    return _theme


def register_theme(name, theme):
    """Register a new or updated theme.

    Parameters
    ----------
    name : str
        Name of requested theme.
    theme : dict of str: str, Theme
        Theme mapping elements to colors.
    """
    if not isinstance(theme, Theme):
        theme = Theme(**theme)
    _themes[name] = theme


def unregister_theme(name):
    """Remove existing theme.

    Parameters
    ----------
    name : str
        Name of the theme to be removed.
    """
    if name in _themes:
        _themes.pop(name)


def available_themes():
    """List available themes.

    Returns
    -------
    list of str
        Names of available themes.
    """
    return tuple(_themes) + ("system",)


def rebuild_theme_settings():
    """update theme information in settings.

    here we simply update the settings to reflect current list of available
    themes.
    """
    from ..settings import get_settings

    settings = get_settings()
    settings.appearance.refresh_themes()


_themes: EventedDict[str, Theme] = EventedDict(
    {
        'dark': Theme(
            **{
                'name': 'dark',
                'background': 'rgb(38, 41, 48)',
                'foreground': 'rgb(65, 72, 81)',
                'primary': 'rgb(90, 98, 108)',
                'secondary': 'rgb(134, 142, 147)',
                'highlight': 'rgb(106, 115, 128)',
                'text': 'rgb(240, 241, 242)',
                'icon': 'rgb(209, 210, 212)',
                'warning': 'rgb(153, 18, 31)',
                'current': 'rgb(0, 122, 204)',
                'syntax_style': 'native',
                'console': 'rgb(0, 0, 0)',
                'canvas': 'black',
            }
        ),
        'light': Theme(
            **{
                'name': 'light',
                'background': 'rgb(239, 235, 233)',
                'foreground': 'rgb(214, 208, 206)',
                'primary': 'rgb(188, 184, 181)',
                'secondary': 'rgb(150, 146, 144)',
                'highlight': 'rgb(163, 158, 156)',
                'text': 'rgb(59, 58, 57)',
                'icon': 'rgb(107, 105, 103)',
                'warning': 'rgb(255, 18, 31)',
                'current': 'rgb(253, 240, 148)',
                'syntax_style': 'default',
                'console': 'rgb(255, 255, 255)',
                'canvas': 'white',
            }
        ),
    },
    basetype=Theme,
)


# this function here instead of plugins._npe2 to avoid circular import
def _install_npe2_themes(_themes):
    import npe2

    for theme in npe2.PluginManager.instance().iter_themes():
        # `theme.type` is dark/light and supplies defaults for keys that
        # are not provided by the plugin
        d = _themes[theme.type].dict()
        d.update(theme.colors.dict(exclude_unset=True))
        _themes[theme.id] = Theme(**d)


_install_npe2_themes(_themes)
_themes.events.added.connect(rebuild_theme_settings)
_themes.events.removed.connect(rebuild_theme_settings)



from napari.components.layerlist import LayerList
from napari.layers import Points
from napari.utils.context import LayerListContextKeys


def test_layerlist_context():
    assert 'layers_selection_count' in LayerListContextKeys.__members__

    ctx = {}
    llc = LayerListContextKeys(ctx)
    assert llc.layers_selection_count == 0
    assert ctx['layers_selection_count'] == 0

    layers = LayerList()

    layers.selection.events.changed.connect(llc.update)
    layers.append(Points())
    assert llc.layers_selection_count == 1
    assert ctx['layers_selection_count'] == 1



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QCheckBox, QComboBox, QLabel, QSlider

from ...utils.colormaps import AVAILABLE_COLORMAPS
from ...utils.translations import trans
from ..utils import qt_signals_blocked
from .qt_layer_controls_base import QtLayerControls


class QtTracksControls(QtLayerControls):
    """Qt view and controls for the Tracks layer.

    Parameters
    ----------
    layer : napari.layers.Tracks
        An instance of a Tracks layer.

    Attributes
    ----------
    grid_layout : qtpy.QtWidgets.QGridLayout
        Layout of Qt widget controls for the layer.
    layer : layers.Tracks
        An instance of a Tracks layer.

    """

    def __init__(self, layer):
        super().__init__(layer)

        # NOTE(arl): there are no events fired for changing checkboxes
        self.layer.events.tail_width.connect(self._on_tail_width_change)
        self.layer.events.tail_length.connect(self._on_tail_length_change)
        self.layer.events.head_length.connect(self._on_head_length_change)
        self.layer.events.properties.connect(self._on_properties_change)
        self.layer.events.colormap.connect(self._on_colormap_change)
        self.layer.events.color_by.connect(self._on_color_by_change)

        # combo box for track coloring, we can get these from the properties
        # keys
        self.color_by_combobox = QComboBox()
        self.color_by_combobox.addItems(self.layer.properties_to_color_by)

        self.colormap_combobox = QComboBox()
        for name, colormap in AVAILABLE_COLORMAPS.items():
            display_name = colormap._display_name
            self.colormap_combobox.addItem(display_name, name)

        # slider for track head length
        self.head_length_slider = QSlider(Qt.Horizontal)
        self.head_length_slider.setFocusPolicy(Qt.NoFocus)
        self.head_length_slider.setMinimum(0)
        self.head_length_slider.setMaximum(self.layer._max_length)
        self.head_length_slider.setSingleStep(1)

        # slider for track tail length
        self.tail_length_slider = QSlider(Qt.Horizontal)
        self.tail_length_slider.setFocusPolicy(Qt.NoFocus)
        self.tail_length_slider.setMinimum(1)
        self.tail_length_slider.setMaximum(self.layer._max_length)
        self.tail_length_slider.setSingleStep(1)

        # slider for track edge width
        self.tail_width_slider = QSlider(Qt.Horizontal)
        self.tail_width_slider.setFocusPolicy(Qt.NoFocus)
        self.tail_width_slider.setMinimum(1)
        self.tail_width_slider.setMaximum(int(2 * self.layer._max_width))
        self.tail_width_slider.setSingleStep(1)

        # checkboxes for display
        self.id_checkbox = QCheckBox()
        self.tail_checkbox = QCheckBox()
        self.tail_checkbox.setChecked(True)
        self.graph_checkbox = QCheckBox()
        self.graph_checkbox.setChecked(True)

        self.tail_width_slider.valueChanged.connect(self.change_tail_width)
        self.tail_length_slider.valueChanged.connect(self.change_tail_length)
        self.head_length_slider.valueChanged.connect(self.change_head_length)
        self.tail_checkbox.stateChanged.connect(self.change_display_tail)
        self.id_checkbox.stateChanged.connect(self.change_display_id)
        self.graph_checkbox.stateChanged.connect(self.change_display_graph)
        self.color_by_combobox.currentTextChanged.connect(self.change_color_by)
        self.colormap_combobox.currentTextChanged.connect(self.change_colormap)

        # grid_layout created in QtLayerControls
        # addWidget(widget, row, column, [row_span, column_span])

        self.grid_layout.addWidget(QLabel(trans._('color by:')), 0, 0)
        self.grid_layout.addWidget(self.color_by_combobox, 0, 1)
        self.grid_layout.addWidget(QLabel(trans._('colormap:')), 1, 0)
        self.grid_layout.addWidget(self.colormap_combobox, 1, 1)
        self.grid_layout.addWidget(QLabel(trans._('blending:')), 2, 0)
        self.grid_layout.addWidget(self.blendComboBox, 2, 1)
        self.grid_layout.addWidget(QLabel(trans._('opacity:')), 3, 0)
        self.grid_layout.addWidget(self.opacitySlider, 3, 1)
        self.grid_layout.addWidget(QLabel(trans._('tail width:')), 4, 0)
        self.grid_layout.addWidget(self.tail_width_slider, 4, 1)
        self.grid_layout.addWidget(QLabel(trans._('tail length:')), 5, 0)
        self.grid_layout.addWidget(self.tail_length_slider, 5, 1)
        self.grid_layout.addWidget(QLabel(trans._('head length:')), 6, 0)
        self.grid_layout.addWidget(self.head_length_slider, 6, 1)
        self.grid_layout.addWidget(QLabel(trans._('tail:')), 7, 0)
        self.grid_layout.addWidget(self.tail_checkbox, 7, 1)
        self.grid_layout.addWidget(QLabel(trans._('show ID:')), 8, 0)
        self.grid_layout.addWidget(self.id_checkbox, 8, 1)
        self.grid_layout.addWidget(QLabel(trans._('graph:')), 9, 0)
        self.grid_layout.addWidget(self.graph_checkbox, 9, 1)
        self.grid_layout.setRowStretch(9, 1)
        self.grid_layout.setColumnStretch(1, 1)
        self.grid_layout.setSpacing(4)

        self._on_tail_length_change()
        self._on_tail_width_change()
        self._on_colormap_change()
        self._on_color_by_change()

    def _on_tail_width_change(self):
        """Receive layer model track line width change event and update slider."""
        with self.layer.events.tail_width.blocker():
            value = int(2 * self.layer.tail_width)
            self.tail_width_slider.setValue(value)

    def _on_tail_length_change(self):
        """Receive layer model track line width change event and update slider."""
        with self.layer.events.tail_length.blocker():
            value = self.layer.tail_length
            self.tail_length_slider.setValue(value)

    def _on_head_length_change(self):
        """Receive layer model track line width change event and update slider."""
        with self.layer.events.head_length.blocker():
            value = self.layer.head_length
            self.head_length_slider.setValue(value)

    def _on_properties_change(self):
        """Change the properties that can be used to color the tracks."""
        with self.layer.events.properties.blocker():

            with qt_signals_blocked(self.color_by_combobox):
                self.color_by_combobox.clear()
            self.color_by_combobox.addItems(self.layer.properties_to_color_by)

    def _on_colormap_change(self):
        """Receive layer model colormap change event and update combobox."""
        with self.layer.events.colormap.blocker():
            self.colormap_combobox.setCurrentIndex(
                self.colormap_combobox.findData(self.layer.colormap)
            )

    def _on_color_by_change(self):
        """Receive layer model color_by change event and update combobox."""
        with self.layer.events.color_by.blocker():
            color_by = self.layer.color_by

            idx = self.color_by_combobox.findText(
                color_by, Qt.MatchFixedString
            )
            self.color_by_combobox.setCurrentIndex(idx)

    def change_tail_width(self, value):
        """Change track line width of shapes on the layer model.

        Parameters
        ----------
        value : float
            Line width of track tails.
        """
        self.layer.tail_width = float(value) / 2.0

    def change_tail_length(self, value):
        """Change edge line backward length of shapes on the layer model.

        Parameters
        ----------
        value : int
            Line length of track tails.
        """
        self.layer.tail_length = value

    def change_head_length(self, value):
        """Change edge line forward length of shapes on the layer model.

        Parameters
        ----------
        value : int
            Line length of track tails.
        """
        self.layer.head_length = value

    def change_display_tail(self, state):
        self.layer.display_tail = self.tail_checkbox.isChecked()

    def change_display_id(self, state):
        self.layer.display_id = self.id_checkbox.isChecked()

    def change_display_graph(self, state):
        self.layer.display_graph = self.graph_checkbox.isChecked()

    def change_color_by(self, value: str):
        self.layer.color_by = value

    def change_colormap(self, colormap: str):
        self.layer.colormap = self.colormap_combobox.currentData()



from typing import TYPE_CHECKING, Dict, Optional, Tuple

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget

from napari._qt.dialogs.qt_reader_dialog import get_reader_choice_for_file


class MockQtReaderDialog:
    """Dialog for user to select a reader plugin for a given file extension or folder"""

    def __init__(
        self,
        pth: str = '',
        parent: 'QWidget' = None,
        readers: Dict[str, str] = {},
        error_message: str = '',
        extension: str = '',
    ):
        self._current_file = pth
        self.readers = readers
        self.error_message = error_message
        self._extension = extension
        self._plugin_choice = None
        self._persist_choice = True
        self._cancelled = False

    def _set_plugin_choice(self, key):
        """Set plugin choice for use in tests"""
        self._plugin_choice = key

    def _set_persist_choice(self, persist: bool):
        """Set persistence checkbox choice for use in tests"""
        self._persist_choice = persist

    def _set_user_cancelled(self):
        """Mock that user chose to cancel on the dialog"""
        self._cancelled = True

    def get_user_choices(self) -> Optional[Tuple[str, bool]]:
        """Mock function for 'executing' dialog and getting user choices"""
        if self._cancelled:
            return

        if len(self.readers) == 1:
            return self._plugin_choice, False

        return self._plugin_choice, self._persist_choice


def test_get_reader_choice_single_reader():
    filename = './my_file.abc'
    readers = {'disp-name': 'plugin_name'}
    dialog = MockQtReaderDialog(filename, None, readers)
    choice = get_reader_choice_for_file(dialog, readers, False)

    assert choice[0] == 'disp-name'
    assert choice[1] is False


def test_get_reader_choice_cancel():
    filename = './my_file.abc'
    readers = {'disp-name': 'plugin_name', 'p1': 'p2'}
    dialog = MockQtReaderDialog(filename, None, readers)
    dialog._set_user_cancelled()

    choice = get_reader_choice_for_file(dialog, readers, False)
    assert choice is None


def test_get_reader_choice_many_persist():
    filename = './my_file.abc'
    readers = {'disp-name': 'plugin_name', 'p1': 'p2'}
    dialog = MockQtReaderDialog(filename, None, readers)
    dialog._set_plugin_choice('p1')

    choice = get_reader_choice_for_file(dialog, readers, False)
    assert choice[0] == 'p1'
    assert choice[1] is True


def test_get_reader_choice_no_persist():
    filename = './my_file.abc'
    readers = {'disp-name': 'plugin_name', 'p1': 'p2'}
    dialog = MockQtReaderDialog(filename, None, readers)
    dialog._set_plugin_choice('p1')
    dialog._set_persist_choice(False)

    choice = get_reader_choice_for_file(dialog, readers, False)
    assert choice[0] == 'p1'
    assert choice[1] is False



from pathlib import Path

from npe2 import PluginManager as _PluginManager
from npe2 import PluginManifest
from npe2.manifest.package_metadata import PackageMetadata

from ..settings import get_settings
from ._plugin_manager import NapariPluginManager

__all__ = ["plugin_manager", "menu_item_template"]

_npe2pm = _PluginManager.instance()

# this is a workaround for the fact that briefcase does not seem to
# include napari's entry_points.txt in the bundled app, so the builtin plugins
# don't get detected.  So we just register it manually.  This could potentially
# be removed when we move to a different bundle strategy
if 'napari' not in _npe2pm._manifests:
    mf_file = Path(__file__).parent.parent / 'builtins.yaml'
    mf = PluginManifest.from_file(mf_file)
    mf.package_metadata = PackageMetadata.for_package('napari')
    _npe2pm.register(mf)

# the main plugin manager instance for the `napari` plugin namespace.
plugin_manager = NapariPluginManager()
plugin_manager._initialize()

# Disable plugins listed as disabled in settings, or detected in npe2
_from_npe2 = {m.package_metadata.name for m in _npe2pm._manifests.values()}
_toblock = get_settings().plugins.disabled_plugins.union(_from_npe2)
plugin_manager._blocked.update(_toblock)

#: Template to use for namespacing a plugin item in the menu bar
menu_item_template = '{}: {}'



import pytest

from napari.utils.context._context_keys import (
    MISSING,
    ContextKey,
    ContextKeyInfo,
    ContextNamespace,
)


def test_context_key_info():
    info = ContextKey.info()
    assert isinstance(info, list) and len(info)
    assert all(isinstance(x, ContextKeyInfo) for x in info)


def _adder(x: list) -> int:
    return sum(x)


def test_context_namespace():
    class Ns(ContextNamespace):
        my_key = ContextKey[list, int](0, 'description', _adder)
        optional_key = ContextKey[None, str](description='might be missing')

    assert 'my_key' in Ns.__members__
    assert str(Ns.my_key) == 'my_key'

    assert any(x.description == 'description' for x in ContextKey.info())
    # make sure the type hints were inferred from adder
    assert Ns.my_key.__orig_class__.__args__ == (list, int)  # type: ignore

    assert isinstance(Ns.my_key, ContextKey)

    ctx = {}
    ns = Ns(ctx)

    assert ns.my_key == 0
    assert ctx['my_key'] == 0
    ns.my_key = 2
    assert ctx['my_key'] == 2

    assert 'optional_key' not in ctx
    assert ns.optional_key is MISSING
    ns.reset('optional_key')  # shouldn't raise error to reset a missing key
    # maybe the key is there though
    ctx['optional_key'] = 'hi'
    assert ns.optional_key == 'hi'

    ns.reset_all()
    assert ctx['my_key'] == 0
    assert 'optional_key' not in ctx
    assert repr(ns) == "{'my_key': 0, 'optional_key': MISSING}"


def test_good_naming():
    with pytest.raises(RuntimeError):
        # you're not allowed to create a key with an id different from
        # it's attribute name
        class Ns(ContextNamespace):
            my_key = ContextKey(id='not_my_key')  # type: ignore



import numpy as np
from vispy.color import Colormap as VispyColormap
from vispy.geometry import MeshData

from ..visuals.surface import SurfaceVisual
from .base import VispyBaseLayer


class VispySurfaceLayer(VispyBaseLayer):
    """Vispy view for the surface layer.

    View is based on the vispy mesh node and uses default values for
    lighting direction and lighting color. More information can be found
    here https://github.com/vispy/vispy/blob/main/vispy/visuals/mesh.py
    """

    def __init__(self, layer):
        node = SurfaceVisual()
        self._meshdata = None
        super().__init__(layer, node)

        self.layer.events.colormap.connect(self._on_colormap_change)
        self.layer.events.contrast_limits.connect(
            self._on_contrast_limits_change
        )
        self.layer.events.gamma.connect(self._on_gamma_change)
        self.layer.events.shading.connect(self._on_shading_change)
        self.layer.wireframe.events.visible.connect(
            self._on_wireframe_visible_change
        )
        self.layer.wireframe.events.width.connect(
            self._on_wireframe_width_change
        )
        self.layer.wireframe.events.color.connect(
            self._on_wireframe_color_change
        )
        self.layer.normals.face.events.connect(self._on_face_normals_change)
        self.layer.normals.vertex.events.connect(
            self._on_vertex_normals_change
        )

        self.reset()
        self._on_data_change()

    def _on_data_change(self):
        if len(self.layer._data_view) == 0 or len(self.layer._view_faces) == 0:
            vertices = None
            faces = None
            vertex_values = np.array([0])
        else:
            # Offsetting so pixels now centered
            # coerce to float to solve vispy/vispy#2007
            # reverse order to get zyx instead of xyz
            vertices = np.asarray(
                self.layer._data_view[:, ::-1], dtype=np.float32
            )
            # due to above xyz>zyx, also reverse order of faces to fix handedness of normals
            faces = self.layer._view_faces[:, ::-1]
            vertex_values = self.layer._view_vertex_values

        if (
            vertices is not None
            and self.layer._ndisplay == 3
            and self.layer.ndim == 2
        ):
            vertices = np.pad(vertices, ((0, 0), (0, 1)))

        # manually detach filters when we go to 2D to avoid dimensionality issues
        # see comments in napari#3475. The filter is set again after set_data!
        if self.layer._ndisplay == 2:
            filt = self.node.shading_filter
            try:
                self.node.detach(filt)
                self.node.shading = None
                self.node.shading_filter = None
            except ValueError:
                # sometimes we try to detach non-attached filters, which causes a ValueError
                pass

        self.node.set_data(
            vertices=vertices, faces=faces, vertex_values=vertex_values
        )

        # disable normals in 2D to avoid shape errors
        if self.layer._ndisplay == 2:
            meshdata = MeshData()
        else:
            meshdata = self.node.mesh_data
        self._meshdata = meshdata

        self._on_face_normals_change()
        self._on_vertex_normals_change()
        self._on_shading_change()

        self.node.update()
        # Call to update order of translation values with new dims:
        self._on_matrix_change()

    def _on_colormap_change(self):
        if self.layer.gamma != 1:
            # when gamma!=1, we instantiate a new colormap with 256 control
            # points from 0-1
            colors = self.layer.colormap.map(
                np.linspace(0, 1, 256) ** self.layer.gamma
            )
            cmap = VispyColormap(colors)
        else:
            cmap = VispyColormap(*self.layer.colormap)
        if self.layer._ndisplay == 3:
            self.node.view_program['texture2D_LUT'] = (
                cmap.texture_lut() if (hasattr(cmap, 'texture_lut')) else None
            )
        self.node.cmap = cmap

    def _on_contrast_limits_change(self):
        self.node.clim = self.layer.contrast_limits

    def _on_gamma_change(self):
        self._on_colormap_change()

    def _on_shading_change(self):
        shading = None if self.layer.shading == 'none' else self.layer.shading
        if self.layer._ndisplay == 3:
            self.node.shading = shading
        self.node.update()

    def _on_wireframe_visible_change(self):
        self.node.wireframe_filter.enabled = self.layer.wireframe.visible
        self.node.update()

    def _on_wireframe_width_change(self):
        self.node.wireframe_filter.width = self.layer.wireframe.width
        self.node.update()

    def _on_wireframe_color_change(self):
        self.node.wireframe_filter.color = self.layer.wireframe.color
        self.node.update()

    def _on_face_normals_change(self):
        self.node.face_normals.visible = self.layer.normals.face.visible
        if self.node.face_normals.visible:
            self.node.face_normals.set_data(
                self._meshdata,
                length=self.layer.normals.face.length,
                color=self.layer.normals.face.color,
                width=self.layer.normals.face.width,
                primitive='face',
            )

    def _on_vertex_normals_change(self):
        self.node.vertex_normals.visible = self.layer.normals.vertex.visible
        if self.node.vertex_normals.visible:
            self.node.vertex_normals.set_data(
                self._meshdata,
                length=self.layer.normals.vertex.length,
                color=self.layer.normals.vertex.color,
                width=self.layer.normals.vertex.width,
                primitive='vertex',
            )

    def reset(self, event=None):
        super().reset()
        self._on_colormap_change()
        self._on_contrast_limits_change()
        self._on_shading_change()
        self._on_wireframe_visible_change()
        self._on_wireframe_width_change()
        self._on_wireframe_color_change()
        self._on_face_normals_change()
        self._on_vertex_normals_change()



import pytest

from ..interactions import Shortcut


@pytest.mark.parametrize(
    'shortcut,reason',
    [
        ('Ctrl-A', 'Ctrl instead of Control'),
        ('Ctrl+A', '+ instead of -'),
        ('Ctrl-AA', 'AA make no sens'),
        ('BB', 'BB make no sens'),
    ],
)
def test_shortcut_invalid(shortcut, reason):

    with pytest.warns(UserWarning):
        Shortcut(shortcut)  # Should be Control-A


def test_minus_shortcut():
    """
    Misc tests minus is properly handled as it is the delimiter
    """
    assert str(Shortcut('-')) == '-'
    assert str(Shortcut('Control--')).endswith('-')
    assert str(Shortcut('Shift--')).endswith('-')


def test_shortcut_qt():

    assert Shortcut('Control-A').qt == 'Control+A'



import os

import pytest

from napari.plugins.io import save_layers

BUILTINS = 'napari'


# the layer_data_and_types fixture is defined in napari/conftest.py
def test_save_layer_single_named_plugin(tmpdir, layer_data_and_types):
    """Test saving a single layer with a named plugin."""
    layers, _, _, filenames = layer_data_and_types

    for layer, fn in zip(layers, filenames):
        path = os.path.join(tmpdir, fn)

        # Check file does not exist
        assert not os.path.isfile(path)

        # Write data
        save_layers(path, [layer], plugin=BUILTINS)

        # Check file now exists
        assert os.path.isfile(path)


# the layer_data_and_types fixture is defined in napari/conftest.py
def test_save_layer_no_results(tmpdir):
    """Test no layers is not an error, and warns on no results."""

    with pytest.warns(UserWarning):
        result = save_layers('no_layers', [], plugin=BUILTINS)
        assert result == []


# the layer_data_and_types fixture is defined in napari/conftest.py
def test_save_layer_single_no_named_plugin(tmpdir, layer_data_and_types):
    """Test saving a single layer without naming plugin."""
    # make writer builtin plugins get called first
    from napari.plugins import plugin_manager

    plugin_manager.hooks.napari_write_image.bring_to_front(['builtins'])
    plugin_manager.hooks.napari_write_points.bring_to_front(['builtins'])

    layers, _, _, filenames = layer_data_and_types

    for layer, fn in zip(layers, filenames):
        path = os.path.join(tmpdir, fn)

        # Check file does not exist
        assert not os.path.isfile(path)

        # Write data
        save_layers(path, [layer])

        # Check file now exists
        assert os.path.isfile(path)


# the layer_data_and_types fixture is defined in napari/conftest.py
def test_save_layer_multiple_named_plugin(tmpdir, layer_data_and_types):
    """Test saving multiple layers with a named plugin."""
    layers, _, _, filenames = layer_data_and_types

    path = os.path.join(tmpdir, 'layers_folder')

    # Check file does not exist
    assert not os.path.isdir(path)

    # Write data
    save_layers(path, layers, plugin=BUILTINS)

    # Check folder now exists
    assert os.path.isdir(path)

    # Check individual files now exist
    for f in filenames:
        assert os.path.isfile(os.path.join(path, f))

    # Check no additional files exist
    assert set(os.listdir(path)) == set(filenames)
    assert set(os.listdir(tmpdir)) == {'layers_folder'}


# the layer_data_and_types fixture is defined in napari/conftest.py
def test_save_layer_multiple_no_named_plugin(tmpdir, layer_data_and_types):
    """Test saving multiple layers without naming a plugin."""
    # make writer builtin plugins get called first
    from napari.plugins import plugin_manager

    plugin_manager.hooks.napari_get_writer.bring_to_front(['builtins'])

    layers, _, _, filenames = layer_data_and_types

    path = os.path.join(tmpdir, 'layers_folder')

    # Check file does not exist
    assert not os.path.isdir(path)

    # Write data
    save_layers(path, layers, plugin=BUILTINS)

    # Check folder now exists
    assert os.path.isdir(path)

    # Check individual files now exist
    for f in filenames:
        assert os.path.isfile(os.path.join(path, f))

    # Check no additional files exist
    assert set(os.listdir(path)) == set(filenames)
    assert set(os.listdir(tmpdir)) == {'layers_folder'}



"""chunk module"""
from ._loader import chunk_loader, synchronous_loading, wait_for_async
from ._request import ChunkLocation, ChunkRequest, LayerRef, OctreeLocation

__all__ = [
    'ChunkLocation',
    'OctreeLocation',
    'ChunkRequest',
    'LayerRef',
    'chunk_loader',
    'wait_for_async',
    'synchronous_loading',
]



"""Performance Monitoring.

The perfmon module lets you instrument your code and visualize its run-time
behavior and timings in Chrome's Tracing GUI.

To enable perfmon define the env var NAPARI_PERFMON as follows:

NAPARI_PERFMON=1
    Activates perfmon, trace using Debug -> Performance Trace menu.

NAPARI_PERFMON=/path/to/config.json
    Configure perfmon using the config.json configuration. See the
    PerfmonConfig docs for the spec of the config file.

Chrome Tracing
---------------

Chrome has a nice built-in performance tool called chrome://tracing. Chrome
can record traces of web applications. But the format is well-documented and
anyone can create the files and use the nice GUI. And other programs accept
the format including:
1) https://www.speedscope.app/ which does flamegraphs (Chrome doesn't).
2) Qt Creator's performance tools.

Monkey Patching
---------------

The best way to add perf_timers is using the perfmon config file. You can
list which methods or functions you want to time, and a perf_timer will be
monkey-patched into each callable on startup. The monkey patching
is done only if perfmon is enabled.

Trace On Start
---------------
Add a line to the config file like:

    "trace_file_on_start": "/Path/to/my/trace.json"

Perfmon will start tracing on startup. You must quit napari with the Quit
command for napari to write trace file. See PerfmonConfig docs.

Manual Timing
-------------

You can also manually add "perf_timer" context objects and
"add_counter_event()" and "add_instant_event()" functions to your code. All
three of these should be removed before merging the PR into main. While
they have almost zero overhead when perfmon is disabled, it's still better
not to leave them in the code. Think of them as similar to debug prints.
"""
import os

from ._config import perf_config
from ._event import PerfEvent
from ._timers import (
    add_counter_event,
    add_instant_event,
    block_timer,
    perf_timer,
    timers,
)

USE_PERFMON = os.getenv("NAPARI_PERFMON", "0") != "0"



import json
from itertools import cycle

import numpy as np
import pytest

from napari.utils.colormaps.categorical_colormap import CategoricalColormap


def test_default_categorical_colormap():
    cmap = CategoricalColormap()
    assert cmap.colormap == {}

    color_cycle = cmap.fallback_color
    np.testing.assert_almost_equal(color_cycle.values, [[1, 1, 1, 1]])
    np.testing.assert_almost_equal(next(color_cycle.cycle), [1, 1, 1, 1])


def test_categorical_colormap_direct():
    """Test a categorical colormap with a provided mapping"""
    colormap = {'hi': np.array([1, 1, 1, 1]), 'hello': np.array([0, 0, 0, 0])}
    cmap = CategoricalColormap(colormap=colormap)

    color = cmap.map(['hi'])
    np.testing.assert_allclose(color, [[1, 1, 1, 1]])
    color = cmap.map(['hello'])
    np.testing.assert_allclose(color, [[0, 0, 0, 0]])

    # test that the default fallback color (white) is applied
    new_color_0 = cmap.map(['not a key'])
    np.testing.assert_almost_equal(new_color_0, [[1, 1, 1, 1]])
    new_cmap = cmap.colormap
    np.testing.assert_almost_equal(new_cmap['not a key'], [1, 1, 1, 1])

    # set a cycle of fallback colors
    new_fallback_colors = [[1, 0, 0, 1], [0, 1, 0, 1]]
    cmap.fallback_color = new_fallback_colors
    new_color_1 = cmap.map(['new_prop 1'])
    np.testing.assert_almost_equal(
        np.squeeze(new_color_1), new_fallback_colors[0]
    )
    new_color_2 = cmap.map(['new_prop 2'])
    np.testing.assert_almost_equal(
        np.squeeze(new_color_2), new_fallback_colors[1]
    )


def test_categorical_colormap_cycle():
    color_cycle = [[1, 1, 1, 1], [1, 0, 0, 1]]
    cmap = CategoricalColormap(fallback_color=color_cycle)

    # verify that no mapping between prop value and color has been set
    assert cmap.colormap == {}

    # the values used to create the color cycle can be accessed via fallback color
    np.testing.assert_almost_equal(cmap.fallback_color.values, color_cycle)

    # map 2 colors, verify their colors are returned in order
    colors = cmap.map(['hi', 'hello'])
    np.testing.assert_almost_equal(colors, color_cycle)

    # map a third color and verify the colors wrap around
    third_color = cmap.map(['bonjour'])
    np.testing.assert_almost_equal(np.squeeze(third_color), color_cycle[0])


def test_categorical_colormap_cycle_as_dict():
    color_values = np.array([[1, 1, 1, 1], [1, 0, 0, 1]])
    color_cycle = cycle(color_values)
    fallback_color = {'values': color_values, 'cycle': color_cycle}
    cmap = CategoricalColormap(fallback_color=fallback_color)

    # verify that no mapping between prop value and color has been set
    assert cmap.colormap == {}

    # the values used to create the color cycle can be accessed via fallback color
    np.testing.assert_almost_equal(cmap.fallback_color.values, color_values)
    np.testing.assert_almost_equal(
        next(cmap.fallback_color.cycle), color_values[0]
    )


fallback_colors = np.array([[1, 0, 0, 1], [0, 1, 0, 1]])


def test_categorical_colormap_from_array():

    cmap = CategoricalColormap.from_array(fallback_colors)
    np.testing.assert_almost_equal(cmap.fallback_color.values, fallback_colors)


color_mapping = {
    'typeA': np.array([1, 1, 1, 1]),
    'typeB': np.array([1, 0, 0, 1]),
}
default_fallback_color = np.array([[1, 1, 1, 1]])


@pytest.mark.parametrize(
    'params,expected',
    [
        ({'colormap': color_mapping}, (color_mapping, default_fallback_color)),
        (
            {'colormap': color_mapping, 'fallback_color': fallback_colors},
            (color_mapping, fallback_colors),
        ),
        ({'fallback_color': fallback_colors}, ({}, fallback_colors)),
        (color_mapping, (color_mapping, default_fallback_color)),
    ],
)
def test_categorical_colormap_from_dict(params, expected):
    cmap = CategoricalColormap.from_dict(params)
    np.testing.assert_equal(cmap.colormap, expected[0])
    np.testing.assert_almost_equal(cmap.fallback_color.values, expected[1])


def test_categorical_colormap_equality():
    color_cycle = [[1, 1, 1, 1], [1, 0, 0, 1]]
    cmap_1 = CategoricalColormap(fallback_color=color_cycle)
    cmap_2 = CategoricalColormap(fallback_color=color_cycle)
    cmap_3 = CategoricalColormap(fallback_color=[[1, 1, 1, 1], [1, 1, 0, 1]])
    cmap_4 = CategoricalColormap(
        colormap={0: np.array([0, 0, 0, 1])}, fallback_color=color_cycle
    )
    assert cmap_1 == cmap_2
    assert cmap_1 != cmap_3
    assert cmap_1 != cmap_4

    # test equality against a different type
    assert cmap_1 != color_cycle


@pytest.mark.parametrize(
    'params',
    [
        {'colormap': color_mapping},
        {'colormap': color_mapping, 'fallback_color': fallback_colors},
        {'fallback_color': fallback_colors},
    ],
)
def test_categorical_colormap_serialization(params):
    cmap_1 = CategoricalColormap(**params)
    cmap_json = cmap_1.json()

    json_dict = json.loads(cmap_json)
    cmap_2 = CategoricalColormap(**json_dict)
    assert cmap_1 == cmap_2



"""This module provides the  :class:`Expr` and its subclasses.

`Expr` is a subclass of `ast.AST` that provides rich dunder methods that
facilitate joining and comparing typed expressions.  It only implements a
subset of ast Expressions (for safety of evaluation), but provides more
than ast.literal_eval.

Expressions that are supported:
- Names: 'myvar' (these must be evaluated along with some context)
- Constants: '1'
- Comparisons: 'myvar > 1'
- Boolean Operators: 'myvar and yourvar'
- Binary Operators: 'myvar + 42'
- Unary Operators: 'not myvar'

Things that are *NOT* supported:
- attribute access: 'my.attr'
- calls: 'f(x)'
- containers (lists, tuples, sets, dicts)
- indexing or slicing
- joined strings (f-strings)
- named expressions (walrus operator)
- comprehensions (list, set, dict, generator)
- statements & assignments (e.g. 'a = b')
"""
from __future__ import annotations

import ast
import sys
from typing import (
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from ..translations import trans

ConstType = Union[None, str, bytes, bool, int, float]
PassedType = TypeVar(
    "PassedType",
    bound=Union[
        ast.cmpop, ast.operator, ast.boolop, ast.unaryop, ast.expr_context
    ],
)
T = TypeVar('T')
T2 = TypeVar('T2', bound=Union[ConstType, 'Expr'])
V = TypeVar('V', bound=ConstType)


def parse_expression(expr: str) -> Expr:
    """Parse string expression into an :class:`Expr` instance.

    Parameters
    ----------
    expr : str
        Expression to parse.

    Returns
    -------
    Expr
        Instance of `Expr`.

    Raises
    ------
    SyntaxError
        If the provided string is not an expression (e.g. it's a statement), or
        if it uses any forbidden syntax components (e.g. Call, Attribute,
        Containers, Indexing, Slicing, f-strings, named expression,
        comprehensions.)
    """
    try:
        # mode='eval' means the expr must consist of a single expression
        tree = ast.parse(expr, mode='eval')
        if not isinstance(tree, ast.Expression):
            raise SyntaxError  # pragma: no cover
        return ExprTranformer().visit(tree.body)
    except SyntaxError as e:
        raise SyntaxError(
            trans._(
                "{expr} is not a valid expression: ({error}).",
                deferred=True,
                expr=f"{expr!r}",
                error=e,
            )
        ) from None


def safe_eval(expr: str, context: Mapping = {}) -> Any:
    """Safely evaluate `expr` string given `context` dict.

    This lets you evaluate a string expression with broader expression
    support than `ast.literal_eval`, but much less support than `eval()`.
    """
    return parse_expression(expr).eval(context)


class Expr(ast.AST, Generic[T]):
    """Base Expression class providing dunder and convenience methods.

    This class is not meant to be instantiated.
    Instead, use :func:`parse_expression`, or the `Expr.parse` classmethod
    to create an expression instance.

    Once created, an expression can be joined with other expressions, or
    constants.

    Methods
    -------
    Outside of the dunder methods providing operator support, key methods are:

    eval(context: dict) -> Any
        Evaluate this expression in the `context` namespace.
    parse(expr: str) -> Expr
        (classmethod) - Parse a string to create a new expression
    _serialize() -> str
        serialize the expression to a string

    Examples
    --------
    >>> expr = parse_expression('myvar > 5')

    # combine expressions with operators
    >>> new_expr = expr & parse_expression('v2')

    # nice repr
    >>> new_expr
    BoolOp(
        op=And(),
        values=[
            Compare(
            left=Name(id='myvar', ctx=Load()),
            ops=[
                Gt()],
            comparators=[
                Constant(value=5)]),
            Name(id='v2', ctx=Load())])

    # evaluate in some context
    >>> new_expr.eval(dict(v2='hello!', myvar=8))
    'hello!'

    # serialize
    >>> print(new_expr)
    myvar > 5 and v2

    One reason you might want to use this is to capture named expressions
    that can be evaluated repeatedly as some underlying context changes.

    .. code-block:: python

        light_is_green = Name[bool]('light_is_green')
        count = Name[int]('count')
        is_ready = light_is_green & count > 5

        assert is_ready.eval({'count': 4, 'light_is_green': True}) == False
        assert is_ready.eval({'count': 7, 'light_is_green': False}) == False
        assert is_ready.eval({'count': 7, 'light_is_green': True}) == True

    this will also preserve type information:
    `reveal_type(is_ready())`  # revealed type is `bool`
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if type(self).__name__ == 'Expr':
            raise RuntimeError(
                trans._(
                    "Don't instantiate Expr. Use `Expr.parse`", deferred=True
                )
            )
        super().__init__(*args, **kwargs)
        ast.fix_missing_locations(self)

    def eval(self, context: Mapping = {}) -> T:
        """Evaluate this expression with names in `context`"""
        code = compile(ast.Expression(body=self), '<Expr>', 'eval')
        try:
            return eval(code, {}, context)
        except NameError:
            miss = {k for k in _iter_names(self) if k not in context}
            raise NameError(
                trans._(
                    'Names required to eval this expression are missing: {miss}',
                    deferred=True,
                    miss=miss,
                )
            )

    @classmethod
    def parse(cls, expr: str) -> Expr:
        """Parse string into Expr (classmethod).

        see docstring of :func:`parse_expression` for details.
        """
        return parse_expression(str(expr))

    def __str__(self) -> str:
        """Serialize this expression to string form."""
        return self._serialize()

    def _serialize(self) -> str:
        """Serialize this expression to string form."""
        return str(ExprSerializer(self))

    def __repr__(self) -> str:
        if sys.version_info >= (3, 9):
            return ast.dump(self, indent=2)
        return ast.dump(self)

    @staticmethod
    def _cast(obj: Any) -> Expr:
        """Cast object into an Expression."""
        return obj if isinstance(obj, Expr) else Constant(obj)

    # boolean operators
    # '&' and '|' are normaly binary operators... but we use them here to
    # combine expression objects meaning "and" and "or".
    # if you want the binary operators, use Expr.bitand, and Expr.bitor

    def __and__(
        self, other: Union[Expr[T2], Expr[T], ConstType, Compare]
    ) -> BoolOp[Union[T, T2]]:
        return BoolOp(ast.And(), [self, other])

    def __or__(
        self, other: Union[Expr[T2], Expr[T], ConstType, Compare]
    ) -> BoolOp[Union[T, T2]]:
        return BoolOp(ast.Or(), [self, other])

    # comparisons

    def __lt__(self, other: Any) -> Compare:
        return Compare(self, [ast.Lt()], [other])

    def __le__(self, other: Any) -> Compare:
        return Compare(self, [ast.LtE()], [other])

    def __eq__(self, other: Any) -> Compare:  # type: ignore
        return Compare(self, [ast.Eq()], [other])

    def __ne__(self, other: Any) -> Compare:  # type: ignore
        return Compare(self, [ast.NotEq()], [other])

    def __gt__(self, other: Any) -> Compare:
        return Compare(self, [ast.Gt()], [other])

    def __ge__(self, other: Any) -> Compare:
        return Compare(self, [ast.GtE()], [other])

    # using __contains__ always returns a bool... so we provide our own
    # Expr.in_ and Expr.not_in methods

    def in_(self, other: Any) -> Compare:
        # not a dunder, use with Expr.in_(a, other)
        return Compare(self, [ast.In()], [other])

    def not_in(self, other: Any) -> Compare:
        # not a dunder, use with Expr.not_in(a, other)
        return Compare(self, [ast.NotIn()], [other])

    # binary operators
    # (note that __and__ and __or__ are reserved for boolean operators.)

    def __add__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.Add(), other)

    def __sub__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.Sub(), other)

    def __mul__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.Mult(), other)

    def __truediv__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.Div(), other)

    def __floordiv__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.FloorDiv(), other)

    def __mod__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.Mod(), other)

    def __matmul__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.MatMult(), other)

    def __pow__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.Pow(), other)

    def __xor__(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.BitXor(), other)

    def bitand(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.BitAnd(), other)

    def bitor(self, other: Union[T, Expr[T]]) -> BinOp[T]:
        return BinOp(self, ast.BitOr(), other)

    # unary operators

    def __neg__(self) -> UnaryOp[T]:
        return UnaryOp(ast.USub(), self)

    def __pos__(self) -> UnaryOp[T]:
        # usually a no-op
        return UnaryOp(ast.UAdd(), self)

    def __invert__(self) -> UnaryOp[T]:
        # note: we're using the invert operator `~` to mean "not ___"
        return UnaryOp(ast.Not(), self)


class Name(Expr[T], ast.Name):
    """A variable name.

    `id` holds the name as a string.
    """

    def __init__(self, id: str, **kwargs: Any) -> None:
        kwargs['ctx'] = ast.Load()
        super().__init__(id, **kwargs)

    def eval(self, context: Mapping = {}) -> T:
        return super().eval(context=context)


class Constant(Expr[V], ast.Constant):
    """A constant value.

    The `value` attribute contains the Python object it represents.
    types supported: NoneType, str, bytes, bool, int, float
    """

    value: V

    def __init__(self, value: V, **kwargs: Any) -> None:
        _valid_type = (type(None), str, bytes, bool, int, float)
        if not isinstance(value, _valid_type):
            raise TypeError(
                trans._(
                    "Constants must be type: {_valid_type!r}",
                    deferred=True,
                    _valid_type=_valid_type,
                )
            )
        super().__init__(value, **kwargs)


class Compare(Expr[bool], ast.Compare):
    """A comparison of two or more values.

    `left` is the first value in the comparison, `ops` the list of operators,
    and `comparators` the list of values after the first element in the
    comparison.
    """

    def __init__(
        self,
        left: Expr,
        ops: Sequence[ast.cmpop],
        comparators: Sequence[Expr],
        **kwargs,
    ) -> None:
        super().__init__(
            Expr._cast(left),
            ops,
            [Expr._cast(c) for c in comparators],
            **kwargs,
        )


class BinOp(Expr[T], ast.BinOp):
    """A binary operation (like addition or division).

    `op` is the operator, and `left` and `right` are any expression nodes.
    """

    def __init__(
        self,
        left: Union[T, Expr[T]],
        op: ast.operator,
        right: Union[T, Expr[T]],
        **k,
    ) -> None:
        super().__init__(Expr._cast(left), op, Expr._cast(right), **k)


class BoolOp(Expr[T], ast.BoolOp):
    """A boolean operation, ‘or’ or ‘and’.

    `op` is Or or And. `values` are the values involved. Consecutive operations
    with the same operator, such as a or b or c, are collapsed into one node
    with several values.

    This doesn’t include `not`, which is a :class:`UnaryOp`.
    """

    def __init__(
        self,
        op: ast.boolop,
        values: Sequence[Union[ConstType, Expr]],
        **kwargs,
    ):
        super().__init__(op, [Expr._cast(v) for v in values], **kwargs)


class UnaryOp(Expr[T], ast.UnaryOp):
    """A unary operation.

    `op` is the operator, and `operand` any expression node.
    """

    def __init__(self, op: ast.unaryop, operand: Expr, **kwargs) -> None:
        super().__init__(op, Expr._cast(operand), **kwargs)


class IfExp(Expr, ast.IfExp):
    """An expression such as `'a if b else c'`.

    `body` if `test` else `orelse`
    """

    def __init__(self, test: Expr, body: Expr, orelse: Expr, **kwargs) -> None:
        super().__init__(
            Expr._cast(test), Expr._cast(body), Expr._cast(orelse), **kwargs
        )


class ExprTranformer(ast.NodeTransformer):
    """This transformer converts an ast.expr into an :class:`Expr`.

    Examples
    --------
    >>> tree = ast.parse('my_var > 11', mode='eval')
    >>> tree = ExprTranformer().visit(tree)  # transformed
    """

    _SUPPORTED_NODES = frozenset(
        k
        for k, v in globals().items()
        if isinstance(v, type) and issubclass(v, Expr)
    )

    # fmt: off
    @overload
    def visit(self, node: ast.expr) -> Expr: ...  # noqa
    @overload
    def visit(self, node: PassedType) -> PassedType: ...  # noqa
    # fmt: on

    def visit(self, node):
        """Visit a node in the tree, transforming into Expr."""

        if isinstance(
            node,
            (
                ast.cmpop,
                ast.operator,
                ast.boolop,
                ast.unaryop,
                ast.expr_context,
            ),
        ):
            # all operation types just get passed through
            return node

        # filter here for supported expression node types
        type_ = type(node).__name__

        if type_ not in ExprTranformer._SUPPORTED_NODES:
            raise SyntaxError(
                trans._(
                    "Type {type_!r} not supported", deferred=True, type_=type_
                )
            )

        # providing fake lineno and col_offset here rather than using
        # ast.fill_missing_locations for typing purposes
        kwargs: Dict[str, Any] = dict(lineno=1, col_offset=0)

        for name, field in ast.iter_fields(node):
            if isinstance(field, ast.expr):
                kwargs[name] = self.visit(field)
            elif isinstance(field, list):
                kwargs[name] = [self.visit(item) for item in field]
            else:
                kwargs[name] = field
        return globals()[type_](**kwargs)


class ExprSerializer(ast.NodeVisitor):
    """Serializes an :class:`Expr` into a string.

    Examples
    --------
    >>> expr = Expr.parse('a + b == c')
    >>> print(expr)
    'a + b == c'

    or ... using this visitor directly:

    >>> ser = ExprSerializer()
    >>> ser.visit(expr)
    >>> out = "".join(ser.result)
    """

    def __init__(self, node: Optional[Expr] = None) -> None:
        self._result: List[str] = []

        def write(*params):
            for item in params:
                if isinstance(item, ast.AST):
                    self.visit(item)
                elif item:
                    self._result.append(item)

        self.write = write

        if node is not None:
            self.visit(node)

    def __str__(self):
        return "".join(self._result)

    def visit_Name(self, node: ast.Name) -> None:
        self.write(node.id)

    def visit_Constant(self, node: ast.Constant) -> None:
        self.write(repr(node.value))

    def visit_BoolOp(self, node: ast.BoolOp) -> Any:
        op = f' {_OPS[type(node.op)]} '
        for idx, value in enumerate(node.values):
            self.write(idx and op or '', value)

    def visit_Compare(self, node: ast.Compare) -> None:
        self.visit(node.left)
        for op, right in zip(node.ops, node.comparators):
            self.write(f' {_OPS[type(op)]} ', right)

    def visit_BinOp(self, node: ast.BinOp) -> None:
        self.write(node.left, f' {_OPS[type(node.op)]} ', node.right)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> None:
        sym = _OPS[type(node.op)]
        self.write(sym, ' ' if sym.isalpha() else '', node.operand)

    def visit_IfExp(self, node: ast.IfExp) -> Any:
        self.write(node.body, ' if ', node.test, ' else ', node.orelse)


OpType = Union[
    Type[ast.operator], Type[ast.cmpop], Type[ast.boolop], Type[ast.unaryop]
]
_OPS: Dict[OpType, str] = {
    # ast.boolop
    ast.Or: 'or',
    ast.And: 'and',
    # ast.cmpop
    ast.Eq: '==',
    ast.Gt: '>',
    ast.GtE: '>=',
    ast.In: 'in',
    ast.Is: 'is',
    ast.NotEq: '!=',
    ast.Lt: '<',
    ast.LtE: '<=',
    ast.NotIn: 'not in',
    ast.IsNot: 'is not',
    # ast.operator
    ast.BitOr: '|',
    ast.BitXor: '^',
    ast.BitAnd: '&',
    ast.LShift: '<<',
    ast.RShift: '>>',
    ast.Add: '+',
    ast.Sub: '-',
    ast.Mult: '*',
    ast.Div: '/',
    ast.Mod: '%',
    ast.FloorDiv: '//',
    ast.MatMult: '@',
    ast.Pow: '**',
    # ast.unaryop
    ast.Not: 'not',
    ast.Invert: '~',
    ast.UAdd: '+',
    ast.USub: '-',
}


def _iter_names(expr: Expr) -> Iterator[str]:
    """Iterate all (nested) names used in the expression.

    Could be used to provide nicer error messages when eval() fails.
    """
    if isinstance(expr, Name):
        yield expr.id
    elif isinstance(expr, Expr):
        for field, val in ast.iter_fields(expr):
            val = val if isinstance(val, list) else [val]
            for v in val:
                yield from _iter_names(v)



import sys
from enum import auto

from ...utils.misc import StringEnum
from ._shapes_models import Ellipse, Line, Path, Polygon, Rectangle


class Mode(StringEnum):
    """Mode: Interactive mode. The normal, default mode is PAN_ZOOM, which
    allows for normal interactivity with the canvas.

    The SELECT mode allows for entire shapes to be selected, moved and
    resized.

    The DIRECT mode allows for shapes to be selected and their individual
    vertices to be moved.

    The VERTEX_INSERT and VERTEX_REMOVE modes allow for individual
    vertices either to be added to or removed from shapes that are already
    selected. Note that shapes cannot be selected in this mode.

    The ADD_RECTANGLE, ADD_ELLIPSE, ADD_LINE, ADD_PATH, and ADD_POLYGON
    modes all allow for their corresponding shape type to be added.
    """

    PAN_ZOOM = auto()
    SELECT = auto()
    DIRECT = auto()
    ADD_RECTANGLE = auto()
    ADD_ELLIPSE = auto()
    ADD_LINE = auto()
    ADD_PATH = auto()
    ADD_POLYGON = auto()
    VERTEX_INSERT = auto()
    VERTEX_REMOVE = auto()
    TRANSFORM = auto()


class ColorMode(StringEnum):
    """
    ColorMode: Color setting mode.

    DIRECT (default mode) allows each shape to be set arbitrarily

    CYCLE allows the color to be set via a color cycle over an attribute

    COLORMAP allows color to be set via a color map over an attribute
    """

    DIRECT = auto()
    CYCLE = auto()
    COLORMAP = auto()


class Box:
    """Box: Constants associated with the vertices of the interaction box"""

    WITH_HANDLE = [0, 1, 2, 3, 4, 5, 6, 7, 9]
    LINE_HANDLE = [7, 6, 4, 2, 0, 7, 8]
    LINE = [0, 2, 4, 6, 0]
    TOP_LEFT = 0
    TOP_CENTER = 7
    LEFT_CENTER = 1
    BOTTOM_RIGHT = 4
    BOTTOM_LEFT = 2
    CENTER = 8
    HANDLE = 9
    LEN = 8


BACKSPACE = 'delete' if sys.platform == 'darwin' else 'backspace'


class ShapeType(StringEnum):
    """ShapeType: Valid shape type."""

    RECTANGLE = auto()
    ELLIPSE = auto()
    LINE = auto()
    PATH = auto()
    POLYGON = auto()


shape_classes = {
    ShapeType.RECTANGLE: Rectangle,
    ShapeType.ELLIPSE: Ellipse,
    ShapeType.LINE: Line,
    ShapeType.PATH: Path,
    ShapeType.POLYGON: Polygon,
}



import ast

import numpy as np
import pytest

from napari.components.layerlist import LayerList
from napari.layers import Image, Labels, Shapes
from napari.layers._layer_actions import (
    _LAYER_ACTIONS,
    ContextAction,
    SubMenu,
    _convert,
    _convert_dtype,
    _project,
)
from napari.utils.context._expressions import Expr
from napari.utils.context._layerlist_context import LayerListContextKeys


class assert_expression_variables(ast.NodeVisitor):
    def __init__(self, expression, names) -> None:
        self._variables: list[str] = []
        self.visit(ast.parse(expression))
        for i in self._variables:
            assert i in names

    def visit_Name(self, node):
        self._variables.append(node.id)


def test_layer_actions():
    """Test that all variables used in layer actions expressions are
    keys in CONTEXT_KEYS.
    """
    names = set(LayerListContextKeys.__members__)
    valid_keys = set.union(
        set(ContextAction.__annotations__), set(SubMenu.__annotations__)
    )
    for action_dict in _LAYER_ACTIONS:
        for action in action_dict.values():
            assert set(action).issubset(valid_keys)
            expr = action.get('enable_when')
            if not expr:
                continue
            assert isinstance(expr, (bool, Expr))
            if isinstance(expr, Expr):
                assert_expression_variables(expr, names)
            expr = action.get('show_when', None)
            if isinstance(expr, Expr):
                assert_expression_variables(expr, names)


@pytest.mark.parametrize(
    'mode', ['max', 'min', 'std', 'sum', 'mean', 'median']
)
def test_projections(mode):
    ll = LayerList()
    ll.append(Image(np.random.rand(8, 8, 8)))
    assert len(ll) == 1
    assert ll[-1].data.ndim == 3
    _project(ll, mode=mode)
    assert len(ll) == 2
    # because we use keepdims = True
    assert ll[-1].data.shape == (1, 8, 8)


@pytest.mark.parametrize(
    'mode',
    ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64'],
)
def test_convert_dtype(mode):
    ll = LayerList()
    data = np.zeros((10, 10), dtype=np.int16)
    ll.append(Labels(data))
    assert ll[-1].data.dtype == np.int16

    data[5, 5] = 1000
    assert data[5, 5] == 1000
    if mode == 'int8' or mode == 'uint8':
        # label value 1000 is outside of the target data type range.
        with pytest.raises(AssertionError):
            _convert_dtype(ll, mode=mode)
        assert ll[-1].data.dtype == np.int16
    else:
        _convert_dtype(ll, mode=mode)
        assert ll[-1].data.dtype == np.dtype(mode)

    assert ll[-1].data[5, 5] == 1000
    assert ll[-1].data.flatten().sum() == 1000


@pytest.mark.parametrize(
    'input, type_',
    [
        (Image(np.random.rand(10, 10)), 'labels'),
        (Labels(np.ones((10, 10), dtype=int)), 'image'),
        (Shapes([np.array([[0, 0], [0, 10], [10, 0], [10, 10]])]), 'labels'),
    ],
)
def test_convert_layer(input, type_):
    ll = LayerList()
    ll.append(input)
    assert ll[0]._type_string != type_
    _convert(ll, type_)
    assert ll[0]._type_string == type_
